from .OMRutils_v6 import *
from .shuffle_tf_and_mcq_v7 import create_set_2_and_set3_and_correct_answer_excel
from .utils_Latex_create_update_question_bank import *
from .utils_Latex_Exam_Questions_Generator import *
from .utils_Scanned_Script_check import *


import re
import random
from docx import Document
from docx.shared import Pt
from copy import deepcopy
from openpyxl import load_workbook
import os
import shutil



def copy_paragraph_with_formatting(source_para, target_doc):
    target_para = target_doc.add_paragraph()
    for run in source_para.runs:
        new_run = target_para.add_run(run.text)
        new_run.bold = run.bold
        new_run.italic = run.italic
        new_run.underline = run.underline
        new_run.font.name = run.font.name
        new_run.font.size = run.font.size if run.font.size else Pt(12)
    return target_para
    
# ---------- T/F Utilities ----------
def extract_tf_questions_between_headers(doc, tf_header_pattern, mcq_header_pattern):
    tf_questions = []
    in_tf_section = False
    for para in doc.paragraphs:
        text = para.text.strip()
        if not text:
            continue
        if not in_tf_section:
            if re.search(tf_header_pattern, text, re.IGNORECASE):
                in_tf_section = True
            continue
        if re.search(mcq_header_pattern, text, re.IGNORECASE):
            break
        tf_questions.append(text)
    return tf_questions

def count_matches(a, b):
    return sum(i == j for i, j in zip(a, b))

def generate_balanced_overlap_sets(original, num_sets=2, num_trials=2000):
    n = len(original)
    best_sets = []
    best_indices = []
    min_max_overlap = float('inf')
    for _ in range(num_trials):
        candidates = []
        indices_list = []
        for _ in range(num_sets):
            permuted_indices = list(range(n))
            while True:
                random.shuffle(permuted_indices)
                shuffled = [original[i] for i in permuted_indices]
                if shuffled != original:
                    break
            candidates.append(shuffled)
            indices_list.append(permuted_indices)
        all_sets = [original] + candidates
        pairwise_overlaps = []
        for i in range(len(all_sets)):
            for j in range(i + 1, len(all_sets)):
                overlap = count_matches(all_sets[i], all_sets[j])
                pairwise_overlaps.append(overlap)
        max_overlap = max(pairwise_overlaps)
        if max_overlap < min_max_overlap:
            min_max_overlap = max_overlap
            best_sets = candidates
            best_indices = indices_list
    return best_sets, best_indices, min_max_overlap

def shuffle_by_index(questions, index_map):
    return [questions[i] for i in index_map]

def write_tf_to_doc(doc, tf_questions,original_doc,tf_header_re):
    header_found = False
    for para in original_doc.paragraphs:
        if re.search(tf_header_re, para.text, re.IGNORECASE):
            copy_paragraph_with_formatting(para, doc)
            doc.add_paragraph()  # Add space after header
            header_found = True
            break
    if not header_found:
        print("Find True/False")

    #doc.add_paragraph()  # spacing
    for i, q in enumerate(tf_questions, 1):
        doc.add_paragraph(f"{i}. {q}")
    doc.add_paragraph()


# ---------- MCQ Utilities ----------
def is_option_paragraph(text):
    return re.match(r'^[a-fA-F][\).]', text.strip())

def extract_mcq_blocks(doc):
    mcq_blocks = []
    i = 0
    while i < len(doc.paragraphs):
        text = doc.paragraphs[i].text.strip()
        if not text:
            i += 1
            continue
        if not is_option_paragraph(text):
            if i + 1 < len(doc.paragraphs) and is_option_paragraph(doc.paragraphs[i + 1].text.strip()):
                block = {
                    'question_idx': i,
                    'question_text': text,
                    'option_indices': [],
                    'option_paragraphs': []
                }
                j = i + 1
                while j < len(doc.paragraphs):
                    opt_text = doc.paragraphs[j].text.strip()
                    if is_option_paragraph(opt_text):
                        block['option_indices'].append(j)
                        block['option_paragraphs'].append(doc.paragraphs[j])
                        j += 1
                    else:
                        break
                mcq_blocks.append(block)
                i = j
            else:
                i += 1
        else:
            i += 1
    return mcq_blocks



def shift_list(lst, shift_by, direction='right'):
    n = len(lst)
    shift_by %= n
    if direction == 'right':
        return lst[-shift_by:] + lst[:-shift_by]
    elif direction == 'left':
        return lst[shift_by:] + lst[:shift_by]
    else:
        raise ValueError("Direction must be 'left' or 'right'")

def contains_special_phrase(option_text):
    # Define normalized target phrases
    phrases = [
        'none of them',
        'none of these',
        'not one of them',
        'all of them',
        'all of the above',
        'all of above'
        'all of these'
    ]

    # Normalize the input text: lowercase, strip, collapse spaces
    cleaned = re.sub(r'\s+', ' ', option_text.strip().lower())

    # Check if any of the phrases exist in the cleaned text
    return any(phrase in cleaned for phrase in phrases)
    
def shuffle_mcq_options(block, original_correct_label, shift_by=1, direction='right'):
    # Concatenate all option texts
    full_text = ' '.join(p.text.strip() for p in block['option_paragraphs'] if p.text.strip())
    parts = re.split(r'(?=([a-fA-F][\).]))', full_text)
    
    if not parts or len(parts) < 3:
        return [full_text], original_correct_label

    options = []
    labels = []
    i = 1
    while i < len(parts):
        label = parts[i]
        content = parts[i + 1].strip() if i + 1 < len(parts) else ''
        content_cleaned = re.sub(r'^[a-fA-F][\).]\s*', '', content)
        labels.append(label[0].lower())
        options.append(content_cleaned)
        i += 2

    try:
        original_index = labels.index(original_correct_label.lower())
    except ValueError:
        # fallback if label is not found
        original_index = 0

    preserve_last = False
    last_option = None
    if len(options) >= 1 and contains_special_phrase(options[-1]):
        preserve_last = True
        last_option = options.pop()
        if original_index == len(options):
            new_correct_label = chr(ord('a') + len(options))
            options.append(last_option)
            result = [f"{chr(ord('a') + idx)}) {opt}" for idx, opt in enumerate(options)]
            return result, new_correct_label

    if len(options) > 1:
        shifted_options = shift_list(options, shift_by, direction)
        total = len(options)
        shift_by_mod = shift_by % total
        if direction == 'right':
            new_index = (original_index + shift_by_mod) % total
        elif direction == 'left':
            new_index = (original_index - shift_by_mod) % total
    else:
        shifted_options = options
        new_index = original_index

    if preserve_last:
        shifted_options.append(last_option)

    result = [f"{chr(ord('a') + idx)}) {opt}" for idx, opt in enumerate(shifted_options)]
    new_correct_label = chr(ord('a') + new_index)
    return result, new_correct_label


def write_mcq_to_doc(doc, mcqs, original_correct_answers, shift_by, direction,original_doc,mcq_header_re, write_options_in_one_line=True):
    header_found = False
    for para in original_doc.paragraphs:
        if re.search(mcq_header_re, para.text, re.IGNORECASE):
            copy_paragraph_with_formatting(para, doc)
            doc.add_paragraph()  # Add space after header
            header_found = True
            break
    if not header_found:
        print("Answer the following questions:")

    new_correct_answers = []
    for i, block in enumerate(mcqs, 1):
        doc.add_paragraph(f"{i}. {block['question_text']}")

        shuffled_options, new_correct_label = shuffle_mcq_options(block, original_correct_answers[i - 1], shift_by, direction)
        new_correct_answers.append(new_correct_label)
        if write_options_in_one_line:
            option_line = '  '.join(shuffled_options)
            doc.add_paragraph(f"   {option_line}")
        else:
            for opt in shuffled_options:
                doc.add_paragraph(f"   {opt}")
        doc.add_paragraph()
    return new_correct_answers

def report_overlap(a, b, label):
    print(f"Overlap {label}: {count_matches(a, b)}")
    



def copy_word_file_with_overwrite(src_path, dest_path):
    if os.path.exists(dest_path):
        os.remove(dest_path)
        print(f"Removed existing file: {dest_path}")

    shutil.copyfile(src_path, dest_path)
    print(f"Copied {src_path} -> {dest_path}")

def write_to_excel(list_of_values,excel_path,course_name,col='A'):
    col_dict={'C':3,'D':4,'E':5,'F':6,'G':7,'J':10,'K':11,'L':12,'M':13,'N':14}
    col_num=col_dict[col]
    # Load the existing Excel workbook
    
    wb = load_workbook(excel_path)
    # Select the active sheet
    ws = wb.active  # Or specify: ws = wb["Sheet1"]


    # Start from row 2
    start_row = 2
    # Write values to a specific column
    for i, answer in enumerate(list_of_values):
        ws.cell(row=start_row + i, column=col_num, value=answer)

    ws.cell(row=start_row, column=1, value=course_name)
    
    # Save workbook
    wb.save(excel_path)
    print("\nValues written to Excel column:", col)



def create_set_2_and_set3_and_correct_answer_excel(base_dir,MS_word_folder_name,course_name,set_1_tf_answers,set_1_mcq_answers,correct_tf_mark,wrong_tf_mark,correct_mcq_mark,wrong_mcq_mark,return_debug=False):
    print("\n\n\n\n")
    print(base_dir)
    print("\n\n\n\n")
    errors = []
    tf_section_errors = []
    qa_section_errors = []
    results = {}
    
#    base_dir = os.path.dirname(__file__)  # Folder where the script is located
    static_dir = os.path.join(base_dir, MS_word_folder_name)

    #Creating the required .docx and .xlsx files
    input_file = os.path.join(static_dir, "Set_1_Uploaded.docx")
    
    output_file_2 = os.path.join(static_dir, "Set_2_"+course_name+".docx")
    output_file_3 = os.path.join(static_dir, "Set_3_"+course_name+".docx")
    template_file_2 = os.path.join(static_dir, "Set_2_blank_template.docx")
    template_file_3 = os.path.join(static_dir, "Set_3_blank_template.docx")

    copy_word_file_with_overwrite(template_file_2, output_file_2)
    copy_word_file_with_overwrite(template_file_3, output_file_3)
    

    src_file_excel = os.path.join(static_dir, "Answer_sheet_template.xlsx")
    excel_path = os.path.join(static_dir, "Correct_Answers_"+course_name+".xlsx")
    shutil.copyfile(src_file_excel, excel_path)

    print("Input file:\n",input_file)
    print("\n\n\n")
    original_doc = Document(input_file)
    doc2 = Document(output_file_2)
    doc3 = Document(output_file_3)
    ###################################################
    
    # ---- True/False ----
    try:
        tf_header_re = r'.*true\s*or\s*false'
        mcq_header_re = r'.*answer.*questions'


        original_tf_questions = extract_tf_questions_between_headers(original_doc, tf_header_re, mcq_header_re)
        
        print("\n\nTotal ",len(original_tf_questions), "T/F questions detected. They are\n\n",original_tf_questions)
        results['no of detected tf'] = {"Set1": len(original_tf_questions)}
        
        required_tf_answers = set_1_tf_answers[:len(original_tf_questions)]
        required_correct_tf_marks=correct_tf_mark[:len(original_tf_questions)]
        required_wrong_tf_marks=wrong_tf_mark[:len(original_tf_questions)]
        
        print("\n\nInside the util function\n")
        print(correct_tf_mark,'\n',wrong_tf_mark,'\n',correct_mcq_mark,'\n',wrong_mcq_mark)
        print(type(correct_tf_mark),'\n',type(wrong_tf_mark),'\n',type(correct_mcq_mark),'\n',type(wrong_mcq_mark))

        if len(required_tf_answers) < len(original_tf_questions) or "" in required_tf_answers:
            tf_section_errors.append(f"{len(original_tf_questions)} TF questions detected, but {len(required_tf_answers)} answers provided (some may be missing or blank).<br>")
        if len(required_correct_tf_marks) < len(original_tf_questions) or "" in correct_tf_mark:
            tf_section_errors.append(f"{len(original_tf_questions)} TF questions detected, but marks for 'Mark if Correct' for all of them are not provided. (some may be missing or blank)<br>.")
        if len(required_wrong_tf_marks) < len(original_tf_questions) or "" in wrong_tf_mark:
            tf_section_errors.append(f"{len(original_tf_questions)} TF questions detected, but marks for 'Mark if Wrong' for all of them are not provided. (some may be missing or blank).<br>")
            
        if tf_section_errors:
            numbered_error_messages = []
            for i, msg in enumerate(tf_section_errors, 1): # Start enumeration from 1
                numbered_error_messages.append(f"{i}# {msg}")
            # Join the numbered messages with <br> for HTML display
            combined_error_message = "<br>".join(numbered_error_messages)
            raise ValueError(combined_error_message)        

        write_to_excel(set_1_tf_answers,excel_path,course_name,col='C')        
        write_to_excel(correct_tf_mark,excel_path,course_name,col='F')
        write_to_excel(wrong_tf_mark,excel_path,course_name,col='G')

        # Generate sets
        (sets, indices, max_overlap) = generate_balanced_overlap_sets(set_1_tf_answers)
        set2_tf_answers, set3_tf_answers = sets
        index_map2, index_map3 = indices

        set2_tf_questions = shuffle_by_index(original_tf_questions, index_map2)
        set3_tf_questions = shuffle_by_index(original_tf_questions, index_map3)

        print("\n\nNew T/F answers Set 2:", set2_tf_answers)
        print("\n\nNew T/F answers Set 3:", set3_tf_answers)
        results['tf_answers'] = {
            "Set1": set_1_tf_answers,
            "Set2": set2_tf_answers,
            "Set3": set3_tf_answers,
                                }
        results['tf_overlap'] = {
            "Set1-Set2": count_matches(set_1_tf_answers, set2_tf_answers),
            "Set1-Set3": count_matches(set_1_tf_answers, set3_tf_answers),
            "Set2-Set3": count_matches(set2_tf_answers, set3_tf_answers),
                                }

        report_overlap(set_1_tf_answers, set2_tf_answers, "Set1-Set2")
        report_overlap(set_1_tf_answers, set3_tf_answers, "Set1-Set3")
        report_overlap(set2_tf_answers, set3_tf_answers, "Set2-Set3")

        
        write_tf_to_doc(doc2,  set2_tf_questions,original_doc,tf_header_re)
        doc2.save(output_file_2)
        print("Save T/F in doc2 success")

        write_to_excel(set2_tf_answers,excel_path,course_name,col='D')
                
        
        write_tf_to_doc(doc3,  set3_tf_questions,original_doc,tf_header_re)
        doc3.save(output_file_3)
        print("Save T/F in doc3 success")
        write_to_excel(set3_tf_answers,excel_path,course_name,col='E')
        

        print(f"Saved changes to {excel_path}")


    except Exception as e:
        errors.append(f"True/False section error:<br> {str(e)}")
        print(f"\n\nERROR: {e}\n\n")
    


    # ---- MCQs ----
    try:       
        #Shifting the MCQ answers randomly to the left or right
        shift_by_2 = random.randint(1, 3)
        direction_2 = random.choice(['left', 'right'])

        shift_by_3 = shift_by_2+1
        direction_3 = direction_2

        mcq_blocks = extract_mcq_blocks(original_doc)
        
        #print(mcq_blocks)
        results['no of detected mcq'] = {"Set1": len(mcq_blocks)}

        required_mcq_answers = set_1_mcq_answers[:len(mcq_blocks)]
        required_correct_mcq_marks=correct_mcq_mark[:len(mcq_blocks)]
        required_wrong_mcq_marks=wrong_mcq_mark[:len(mcq_blocks)]
         

        if len(required_mcq_answers) < len(mcq_blocks) or "" in required_mcq_answers:
            qa_section_errors.append(f"{len(mcq_blocks)} MCQ questions detected, but {len(required_mcq_answers)} answers provided (some may be missing or blank).<br>")
        if len(required_correct_mcq_marks) < len(mcq_blocks) or "" in required_correct_mcq_marks:
            qa_section_errors.append(f"{len(mcq_blocks)} MCQ questions detected, but marks for 'Mark if Correct' for all of them are not provided. (some may be missing or blank)<br>.")
        if len(required_wrong_mcq_marks) < len(mcq_blocks) or "" in required_wrong_mcq_marks:
            qa_section_errors.append(f"{len(mcq_blocks)} MCQ questions detected, but marks for 'Mark if Wrong' for all of them are not provided. (some may be missing or blank)<br>.")
        
        if qa_section_errors:
            numbered_error_messages = []
            for i, msg in enumerate(qa_section_errors, 1): # Start enumeration from 1
                numbered_error_messages.append(f"{i}# {msg}")
            # Join the numbered messages with <br> for HTML display
            combined_error_message = "<br>".join(numbered_error_messages)
            raise ValueError(combined_error_message)        



        
        #print(len(macq_blocks))
        new_mcq_answers_2 = write_mcq_to_doc(doc2, mcq_blocks, set_1_mcq_answers, shift_by_2, direction_2,original_doc,mcq_header_re)
        new_mcq_answers_3 = write_mcq_to_doc(doc3, mcq_blocks, set_1_mcq_answers, shift_by_3, direction_3,original_doc,mcq_header_re)
        
        print("\n\nNew MCQ answers Set 2:", new_mcq_answers_2)
        print("New MCQ answers Set 3:", new_mcq_answers_3)

        results['mcq answers'] = {"Set1": set_1_mcq_answers,
                                  "Set2": new_mcq_answers_2,
                                  "Set3": new_mcq_answers_3
                                 }
        
        write_to_excel(set_1_mcq_answers,excel_path,course_name,col='J')
        write_to_excel(correct_mcq_mark,excel_path,course_name,col='M')
        write_to_excel(wrong_mcq_mark,excel_path,course_name,col='N')
        
        doc2.save(output_file_2)
        print("Save Q/A in doc2 success")
        write_to_excel(new_mcq_answers_2,excel_path,course_name,col='K')

        doc3.save(output_file_3)
        print("Save Q/A in doc3 success")
        write_to_excel(new_mcq_answers_3,excel_path,course_name,col='L')  

        

    except Exception as e:
        errors.append(f"MCQ section error:<br> {str(e)}")
        print(f"\n\nERROR: {e}\n\n")


    return results, errors

    #############################################################################################################
    #############################################################################################################
    #############################################################################################################










###########################################################################
# Latex and question processing related functions
###########################################################################

from fastapi import FastAPI, HTTPException
import subprocess
import os
import random
import string
import re
import numpy as np
import itertools
import shutil
from openpyxl import load_workbook
from pathlib import Path
import re
import html



from pydantic import BaseModel
from typing import List
from fastapi.responses import JSONResponse
###########################################################################

# # -------------------------------
# # Parse LaTeX questions
# # -------------------------------


def remove_comment_lines(text: str) -> str:
    """
    Remove full-line LaTeX comments that start with %
    but keep inline % symbols (like in math or text) intact.
    """
    lines = text.splitlines()
    cleaned_lines = [line for line in lines if not line.strip().startswith('%')]
    return "\n".join(cleaned_lines).strip()


def parse_all_latex_questions_from_question_bank(file_path):
    """Parse LaTeX questions (MCQ + True/False) from question bank."""
    with open(file_path, "r", encoding="utf-8") as f:
        content = f.read()

    # Split all \question blocks
    questions_raw = re.split(r'\\question', content)
    parsed_mcq = []
    parsed_tf = []

    for q in questions_raw:
        q = q.strip()
        if not q:
            continue

        # Extract metadata inside square brackets
        meta_match = re.match(r'\[(.*?)\]', q)
        q_type = "mcq"  # default if not specified
        correct_answer = None
        mark = None

        if meta_match:
            metadata = meta_match.group(1)

            # Parse type, answer, and mark
            type_search = re.search(r'type=(\w+)', metadata)
            if type_search:
                q_type = type_search.group(1).lower()

            ans_search = re.search(r'answer=([A-Za-z])', metadata)
            if ans_search:
                correct_answer = ans_search.group(1)

            mark_search = re.search(r'mark=(\d+)', metadata)
            if mark_search:
                mark = int(mark_search.group(1))

            # Remove metadata from the question text
            q = re.sub(r'^\[.*?\]', '', q).strip()

        # === MCQ type ===
        if q_type == "mcq":
            parts = re.split(r'\\begin{choices}', q)
            if len(parts) < 2:
                continue  # skip malformed MCQs

            question_text = remove_comment_lines(parts[0].strip())
            choices_block = remove_comment_lines(parts[1])
    
            # Extract each \choice line
            choices = re.findall(r'\\choice\s+(.*)', choices_block)

            parsed_mcq.append({
                "type": q_type,
                "question": question_text,
                "choices": choices,
                "answer": correct_answer,
                "mark": mark
            })

        # === True/False type ===
        elif q_type == "tf":
            # 1. Cleanly extract only the question statement. 
            # Since TF questions don't use \begin{choices}, 
            # we take all the remaining content of 'q' as the statement.
            question_text = remove_comment_lines(q.strip())

            # 2. Add an explicit check to strip any remaining \end{choices} or other junk 
            # if the question bank is messy. (Defensive coding)
            question_text = re.sub(r'\\end\{choices\}', '', question_text).strip()
            
            parsed_tf.append({
                "type": q_type,
                "question": question_text,
                "choices": [], # Use an empty list to trigger the 'a) True \quad b) False' display
                "answer": correct_answer,
                "mark": mark
            })

    return parsed_mcq, parsed_tf


def generate_balanced_overlap_sets(selected_tf_set_1, num_additional_sets=2, num_trials=2000):
    # This function shuffles the true/false answers among multiple sets.
    # It does that in such a way that there will be minimum number of overlap (same answer for a given question number) between any two sets.
    
    # Extract the correct answers for Set-1 from the list of dictionaries
    Set_1_tf_correct_answers = [q['answer'] for q in selected_tf_set_1]
    
    n = len(Set_1_tf_correct_answers)

    # 🔹 Prevent infinite loop for trivial cases
    if n == 0:
        # No questions at all → return empty structures
        tf_questions_all_sets = [[] for _ in range(num_additional_sets + 1)]
        best_indices = [[] for _ in range(num_additional_sets)]
        tf_answer_overlap_between_sets = [
            (f"Set-{i}, Set-{j}", 0)
            for i in range(1, num_additional_sets + 2)
            for j in range(i + 1, num_additional_sets + 2)
        ]
        return tf_questions_all_sets, best_indices, tf_answer_overlap_between_sets

    elif n == 1:
        # One question only → replicate same across sets
        best_sets = [Set_1_tf_correct_answers.copy() for _ in range(num_additional_sets)]
        best_indices = [[0] for _ in range(num_additional_sets)]
        tf_answer_overlap_between_sets = [
            (f"Set-{i}, Set-{j}", 1)
            for i in range(1, num_additional_sets + 2)
            for j in range(i + 1, num_additional_sets + 2)
        ]
        tf_questions_all_sets = [selected_tf_set_1] + [
            [selected_tf_set_1[0]] for _ in range(num_additional_sets)
        ]
        return tf_questions_all_sets, best_indices, tf_answer_overlap_between_sets

    
    # -------------------------------------------------------------
    # 1. Helper function: Count matches (Overlap)
    # -------------------------------------------------------------
    def count_matches(a, b):
        return sum(i == j for i, j in zip(a, b))

    # -------------------------------------------------------------
    # 2. Nested function: Find the best permutation of answers
    # -------------------------------------------------------------
    def generate_best_shuffled_set():
        n = len(Set_1_tf_correct_answers)
        best_sets = []
        best_indices = []
        min_max_overlap = float('inf')
        
        for _ in range(num_trials):
            candidates = []
            indices_list = []
            
            for _ in range(num_additional_sets):
                permuted_indices = list(range(n))
                attempts = 0
                MAX_ATTEMPTS = 50  # Set a safe limit
                
                while attempts < MAX_ATTEMPTS:
                    random.shuffle(permuted_indices)
                    shuffled = [Set_1_tf_correct_answers[i] for i in permuted_indices]  
                    if shuffled != Set_1_tf_correct_answers:
                        break
                    attempts += 1
                if attempts == MAX_ATTEMPTS:
                    print(f"Warning: Could not find a unique shuffle after {MAX_ATTEMPTS} attempts. Accepting original set.")
                candidates.append(shuffled)
                indices_list.append(permuted_indices)
                
            all_sets = [Set_1_tf_correct_answers] + candidates
            pairwise_overlaps = []
            
            for i in range(len(all_sets)):
                for j in range(i + 1, len(all_sets)):
                    overlap = count_matches(all_sets[i], all_sets[j])
                    pairwise_overlaps.append(overlap)
                    
            max_overlap = max(pairwise_overlaps)
            
            if max_overlap < min_max_overlap:
                min_max_overlap = max_overlap
                best_sets = candidates
                best_indices = indices_list 
                
        # REMOVED: min_max_overlap from return
        return best_sets, best_indices
    
    # -------------------------------------------------------------
    # 3. Nested function: Compute all pairwise overlaps for the best sets
    # -------------------------------------------------------------
    def compute_overlap_between_sets(best_sets):
        tf_answer_overlap_between_sets = []
        
        for set_pair in itertools.combinations(range(num_additional_sets + 1), 2): 
            
            final_all_sets = [Set_1_tf_correct_answers] + best_sets

            index_a, index_b = set_pair
            set_a = final_all_sets[index_a]
            set_b = final_all_sets[index_b]
            
            overlap_count = count_matches(set_a, set_b)
            
            label_a = f'Set-{index_a + 1}'
            label_b = f'Set-{index_b + 1}'
            
            tf_answer_overlap_between_sets.append((f'{label_a}, {label_b}', overlap_count))
            
        # REMOVED: min_max_overlap from return
        return tf_answer_overlap_between_sets 

    # -------------------------------------------------------------
    # 4. NEW Nested function: Use indices to permute the question list
    # -------------------------------------------------------------
    def create_shuffled_question_sets(best_indices):
        shuffled_tf_sets = []
        
        for permutation_indices in best_indices:
            shuffled_set = [selected_tf_set_1[i] for i in permutation_indices]
            shuffled_tf_sets.append(shuffled_set)
            
        return shuffled_tf_sets


    # -------------------------------------------------------------
    # Main function execution flow:
    # -------------------------------------------------------------
    
    # Step A: Find the best answer permutations
    #print("trying to generate best shuffled set")
    best_sets_answers, best_indices = generate_best_shuffled_set() 
    
    #print("Computing overlap")
    # Step B: Calculate final overlaps using the best answer sets
    tf_answer_overlap_between_sets = compute_overlap_between_sets(best_sets_answers)
    #print("Creating shuffled question sets")
    # Step C: Create the final shuffled question lists (list of dictionaries)
    shuffled_tf_sets = create_shuffled_question_sets(best_indices)
    
    #print("combining with set-1")
    # Combine the Set-1 and the newly created sets (Set-2 and Set-3) in a single list
    tf_questions_all_sets = [selected_tf_set_1] + shuffled_tf_sets
    # Return all required results (removed min_max_overlap)
    return tf_questions_all_sets, best_indices, tf_answer_overlap_between_sets



def generate_shuffled_mcq_sets(selected_mcq_set_1, num_additional_sets=2):
    """
    Generates new MCQ sets by shuffling the order of the choices and
    **correctly updates the 'answer' key** to reflect the new position
    of the correct choice.
    """
    
    # Function to convert 0 -> 'a', 1 -> 'b', etc.
    def index_to_letter(index):
        # We use string.ascii_lowercase for guaranteed a-z
        return string.ascii_lowercase[index]

    mcq_questions_all_sets = [selected_mcq_set_1]
    
    # Total number of sets to generate (excluding Set-1)
    for _ in range(num_additional_sets):
        
        shuffled_set = []
        
        # Iterate through each question in the original Set-1 sequence
        for original_q in selected_mcq_set_1:
            
            new_q = original_q.copy()
            
            # --- 1. Find the content of the correct choice in Set-1 ---
            
            # Convert the answer letter ('a', 'b', etc.) back to an index (0, 1, etc.)
            original_answer_index = ord(original_q['answer']) - ord('a')
            
            # Get the content of the correct choice (e.g., 'Force' or '3 x 10^8 m/s')
            correct_choice_content = original_q['choices'][original_answer_index]
            
            # --- 2. Shuffle the choices ---
            
            choices_to_shuffle = new_q['choices'][:] # Use a copy for safe shuffling
            random.shuffle(choices_to_shuffle)
            new_q['choices'] = choices_to_shuffle
            
            # --- 3. Find the new index of the correct content ---
            
            # Find the new index of the correct content in the shuffled list
            new_answer_index = choices_to_shuffle.index(correct_choice_content)
            
            # --- 4. Update the 'answer' key ---
            
            # Convert the new index back to a letter
            new_q['answer'] = index_to_letter(new_answer_index)
            
            # 5. Add the newly shuffled question to the current set
            shuffled_set.append(new_q)
            
        mcq_questions_all_sets.append(shuffled_set)
        
    return mcq_questions_all_sets



# ###################################################################
# #Following function is responsible for processing the latex code
# ###################################################################



def sanitize_for_latex(text):
    """
    Cleans text for LaTeX safety:
    - Escapes only text-mode specials
    - Does NOT escape inside math mode (\(...\), $...$, \frac, etc.)
    - Wraps comparison symbols (> < =) in math mode if found outside math
    """
    if not text:
        return ""

    text = html.unescape(text).strip()

    # Split text into math and non-math parts
    # We'll detect math segments (anything between \( ... \), $...$, or \[ ... \])
    math_segments = []
    def math_replacer(match):
        math_segments.append(match.group(0))
        return f"@@MATH{len(math_segments)-1}@@"

    text = re.sub(r'(\\\([^\)]*\\\)|\\\[[^\]]*\\\]|\$[^$]*\$)', math_replacer, text)

    # Escape only outside math mode
    specials = {
        '&': r'\&', '%': r'\%', '$': r'\$', '#': r'\#', '_': r'\_',
        '{': r'\{', '}': r'\}', '~': r'\textasciitilde{}',
        '^': r'\textasciicircum{}',
        # NOTE: don't escape backslashes globally anymore!
    }
    for k, v in specials.items():
        text = text.replace(k, v)

    # Wrap comparison symbols (> < =) in math mode (outside math)
    def wrap_symbol(m):
        sym = m.group(0)
        return f'\\({sym}\\)'
    text = re.sub(r'(?<![\\$])([><=])(?!=)', wrap_symbol, text)

    # Restore original math segments
    for i, seg in enumerate(math_segments):
        text = text.replace(f"@@MATH{i}@@", seg)

    # Normalize whitespace
    text = re.sub(r'\s+', ' ', text)

    return text




def create_question_latex_code_and_pdf(
    Department,
    exam_title,
    latex_template_file,
    Question_Name,
    selected_mcq_all_sets=[],
    selected_tf_all_sets=[],
    full_mark=0,
    exam_duration="0 min",
    tf_neg=50,
    mcq_neg=25,
    base_folder=None,  # ✅ new argument
):
    """
    Creates LaTeX exam sets and compiles them into PDFs.
    All output files will be stored under base_folder (passed from main.py).
    """

    # ✅ Default fallback if not provided
    if base_folder is None:
        base_folder = Path.home() / "Documents" / "WebCam_OMR"

    # ---------------------------
    # Utility: Format Question
    # ---------------------------
    def format_question_latex(question_data):
        question_raw = question_data.get("question", "")
        question = sanitize_for_latex(question_raw)

        q_type = question_data.get("type", "mcq").lower()
        choices = [sanitize_for_latex(c) for c in question_data.get("choices", [])]
        mark = question_data.get("mark", "")

        if q_type == "mcq":
            MARK_WIDTH = "0.25in"
            GAP_WIDTH = "0.1in"
            COLUMN_WIDTH = f"\\dimexpr\\linewidth-{MARK_WIDTH}-{GAP_WIDTH}\\relax"

            formatted_choices = " ".join(
                [
                    f"{chr(97+i)}) {choice}"
                    + ("" if i == len(choices) - 1 else " \\quad")
                    for i, choice in enumerate(choices)
                ]
            )

            latex = (
                "\\item \n"
                f"\\begin{{tabular}}[t]{{@{{}}p{{{COLUMN_WIDTH}}}@{{\\hspace{{{GAP_WIDTH}}}}}r@{{}}}}\n"
                f"{question}\n"
                "\\par\\vspace{1ex}\n"
                f"\\noindent {formatted_choices} & \n"
                f"\\makebox[{MARK_WIDTH}][r]{{\\textbf{{{mark}}}}}\n"
                "\\end{tabular}\n"
            )
        else:
            latex = "\\item " + question + "\n"

        return latex

    # ---------------------------
    # Utility: Create and Compile LaTeX
    # ---------------------------
    def create_latex_code_and_run(
        latex_template_file,
        output_tex_file,
        exam_title,
        mcq_image_path,
        selected_mcq,
        selected_tf,
        full_mark,
        exam_duration,
        tf_neg,
        mcq_neg,
    ):
        num_tf_questions = len(selected_tf)
        formatted_mcq = [format_question_latex(q) for q in selected_mcq]
        formatted_tf = [format_question_latex(q) for q in selected_tf]

        with open(latex_template_file, "r", encoding="utf-8") as f:
            latex_content = f.read()

        latex_content = latex_content.replace("%MCQ_PLACEHOLDER%", "".join(formatted_mcq))
        latex_content = latex_content.replace("%TF_PLACEHOLDER%", "".join(formatted_tf))
        latex_content = latex_content.replace("Exam Title Here", exam_title)
        latex_content = latex_content.replace("example-image.png", mcq_image_path.replace("\\", "/"))
        latex_content = latex_content.replace("{placeholder_full_mark}", str(full_mark))
        latex_content = latex_content.replace("{placeholder_time}", exam_duration)
        latex_content = latex_content.replace("tf_neg_placeholder", f"{tf_neg}\\%")
        latex_content = latex_content.replace("mcq_neg_placeholder", f"{mcq_neg}\\%")

        tf_total_value = num_tf_questions * 1
        latex_content = latex_content.replace(
            "(no of tf questions $ \\times $ 1 = value)",
            f"({num_tf_questions} $\\times$ 1 = {tf_total_value})",
        )

        with open(output_tex_file, "w", encoding="utf-8") as f:
            f.write(latex_content)

        print(f"✅ LaTeX file created: {output_tex_file}")

        try:
            tex_dir = os.path.dirname(output_tex_file)
            tex_filename = os.path.basename(output_tex_file)

            subprocess.run(
                ["pdflatex", "-interaction=nonstopmode", tex_filename],
                check=True,
                cwd=tex_dir,
            )
            print(f"✅ Successfully compiled {output_tex_file} and PDF is generated.")
        except subprocess.CalledProcessError:
            print("❌ Error: pdflatex compilation failed. Check LaTeX installation.")

    # ---------------------------
    # Prepare Output Folder
    # ---------------------------
    output_folder = base_folder / "Generated_Exams"
    output_folder.mkdir(parents=True, exist_ok=True)

    # Optional cleanup
    for ext in ("*.tex", "*.pdf", "*.aux", "*.log"):
        for file_path in output_folder.glob(ext):
            try:
                file_path.unlink()
            except PermissionError:
                print(f"❌ Cannot delete {file_path}, it may be in use.")

    # ---------------------------
    # Generate PDFs
    # ---------------------------
    output_pdf_file_names = []
    project_root = Path.cwd()
#    project_root = Path(__file__).parent.parent

    for index, (selected_mcq, selected_tf) in enumerate(
        zip(selected_mcq_all_sets, selected_tf_all_sets)
    ):
        output_tex_file = f"{Question_Name}_Set_{index+1}.tex"
        output_tex_path = output_folder / output_tex_file

        mcq_image_path = project_root / "static" / f"{Department}_Set_{index+1}.png"
        mcq_image_path_str = str(mcq_image_path.resolve())
        
        create_latex_code_and_run(
            latex_template_file,
            str(output_tex_path),
            exam_title,
            mcq_image_path_str,
            selected_mcq,
            selected_tf,
            full_mark,
            exam_duration,
            tf_neg,
            mcq_neg,
        )

        output_pdf_file_names.append(output_tex_file.replace(".tex", ".pdf"))

    return output_folder,output_pdf_file_names




def generate_set_wise_answers_and_marks(questions_all_sets,WA_neg_mark_percentage):
    answers_all_sets = []
    for question_set in questions_all_sets:    
        answers_for_current_set = []    
        selected_CA_marks=[]
        selected_WA_marks=[]
        for q in question_set:
            answers_for_current_set.append(q['answer'])     
            selected_CA_marks.append(float(q['mark']))
            selected_WA_marks.append(-float(np.round(float(q['mark'])*WA_neg_mark_percentage/100,2)))
            
        answers_all_sets.append(answers_for_current_set)
    return answers_all_sets,selected_CA_marks,selected_WA_marks



# ######################################################################
# ######################################################################



def create_correct_answer_excel(
    xlsx_template_file,
    output_xlsx_file,
    course_title,
    tf_answers_all_sets=[],
    selected_tf_CA_marks=[],
    selected_tf_WA_marks=[],
    mcq_answers_all_sets=[],
    selected_mcq_CA_marks=[],
    selected_mcq_WA_marks=[] ):
        
    def write_to_excel(list_of_values,excel_path,col='A',course_name=""):
        col_dict={'C':3,'D':4,'E':5,'F':6,'G':7,'J':10,'K':11,'L':12,'M':13,'N':14}
        col_num=col_dict[col]
        # Load the existing Excel workbook
        
        wb = load_workbook(excel_path)
        # Select the active sheet
        ws = wb.active  # Or specify: ws = wb["Sheet1"]


        # Start from row 2
        start_row = 2
        # Write values to a specific column
        for i, answer in enumerate(list_of_values):
            ws.cell(row=start_row + i, column=col_num, value=answer)

        if course_name != "":
            ws.cell(row=start_row, column=1, value=course_name)
        
        # Save workbook
        wb.save(excel_path)
        print("\nValues written to Excel column:", col)

    shutil.copyfile(xlsx_template_file, output_xlsx_file)
    
    write_to_excel(tf_answers_all_sets[0],output_xlsx_file,col='C')   
    write_to_excel(tf_answers_all_sets[1],output_xlsx_file,col='D')   
    write_to_excel(tf_answers_all_sets[2],output_xlsx_file,col='E')        
    write_to_excel(selected_tf_CA_marks,output_xlsx_file,col='F')
    write_to_excel(selected_tf_WA_marks,output_xlsx_file,col='G')
    
    write_to_excel(mcq_answers_all_sets[0],output_xlsx_file,col='J')
    write_to_excel(mcq_answers_all_sets[1],output_xlsx_file,col='K')
    write_to_excel(mcq_answers_all_sets[2],output_xlsx_file,col='L')  

    write_to_excel(selected_mcq_CA_marks,output_xlsx_file,col='M')
    write_to_excel(selected_mcq_WA_marks,output_xlsx_file,col='N',course_name=course_title)
    



# ######################################################################
# ######################################################################


# ###########################################################################
# # Webpage related functions
# ###########################################################################

# app = FastAPI()

async def load_question_bank(base_folder: Path, course_code: str):
    """
    Loads the LaTeX question bank file based on department and course code.
    The base_folder is provided by main.py.
    """
    question_bank_folder = base_folder / "Question_Bank"
    file_path = question_bank_folder / f"{course_code}.tex"

    # ✅ Ensure directory exists
    question_bank_folder.mkdir(parents=True, exist_ok=True)

    if not file_path.exists():
        raise HTTPException(
            status_code=404,
            detail=f"Question bank file not found in {question_bank_folder}. "
                   f"Copy Question Bank (.tex file) in that directory."
        )

    with open(file_path, "r", encoding="utf-8") as f:
        content = f.read()

    return {"content": content}



# ------------------ Pydantic model for exam generation ------------------
class ExamGenerationRequest(BaseModel):
    department: str
    courseCode: str
    examType: str
    courseTitle: str
    fullMark: int
    duration: int
    mcqNeg: int
    tfNeg: int
    randomSelect: bool
    numTf: int
    numMcq: int
    selectedTfQuestions: List[int]
    selectedTfMarks: List[int]
    selectedMcqQuestions: List[int]
    selectedMcqMarks: List[int]



# ------------------ Exam generation handler ------------------
async def generate_exam(request_data: ExamGenerationRequest, base_folder: Path):
    """
    Handles the exam generation request.
    Generates exam sets and returns status messages.
    Paths are derived from base_folder provided by main.py.
    """
    import traceback
    messages = []
    print("Generate_exam called with base_folder=", base_folder)
    
    try:
        #--------------------------------------------#
        Department = request_data.department
        Course_Code = re.sub(r"\s+", "", request_data.courseCode)  # remove spaces
        Exam_type = request_data.examType
        Course_title = request_data.courseTitle
        full_mark = float(request_data.fullMark)
        exam_duration = str(request_data.duration) + " mins"
        mcq_neg_percentage = int(request_data.mcqNeg)
        tf_neg_percentage = int(request_data.tfNeg)
        
        Random_Selection = request_data.randomSelect
        num_tf_to_select = int(request_data.numTf)
        num_mcq_to_select = int(request_data.numMcq)
        Selected_TF_Questions = request_data.selectedTfQuestions
        Selected_MCQ_Questions = request_data.selectedMcqQuestions
        TF_Marks = request_data.selectedTfMarks
        MCQ_Marks = request_data.selectedMcqMarks

        #--------------------------------------------#
        if Course_title.strip():
            exam_title = f"{Exam_type} on {Course_Code}: {Course_title}"
        else:
            exam_title = f"{Exam_type} on {Course_Code}"

        Question_Name = f"{Exam_type}_{Course_Code}"
        question_bank_fileName = f"{Course_Code}.tex"

        # ✅ All paths now derived from base_folder
        question_bank_folder = base_folder / "Question_Bank"
        question_bank_folder.mkdir(parents=True, exist_ok=True)
        question_bank_file = question_bank_folder / question_bank_fileName
        
        generated_exams_dir = base_folder / "Generated_Exams"
        generated_exams_dir.mkdir(parents=True, exist_ok=True)


        # static assets (e.g. templates) can still come from your working directory
        latex_template_file = Path("static") / "latex_template.tex"
        correct_answers_template_file = Path("static") / "Answer_sheet_template.xlsx"
        correct_answer_excel_FN="Correct_Answers_"+Course_Code+".xlsx"
        correct_answers_output_file = generated_exams_dir / correct_answer_excel_FN
        #correct_answers_output_file = os.path.join(base_folder,"Generated_Exams",correct_answer_excel_FN)
        


        #--------------------------------------------#
        # Load questions
        all_mcq, all_tf = parse_all_latex_questions_from_question_bank(question_bank_file)
        

        #--------------------------------------------#
        # Selection logic
        if Random_Selection:
            if num_mcq_to_select > len(all_mcq):
                messages.append("You are trying to select more MCQ questions than are available in the question bank.")
            if num_tf_to_select > len(all_tf):
                messages.append("You are trying to select more T/F questions than are available in the question bank.")

            selected_mcq = random.sample(all_mcq, min(num_mcq_to_select, len(all_mcq)))
            selected_tf = random.sample(all_tf, min(num_tf_to_select, len(all_tf)))
        else:
            selected_mcq = [all_mcq[i - 1] for i in Selected_MCQ_Questions if 0 < i <= len(all_mcq)]
            selected_tf = [all_tf[i - 1] for i in Selected_TF_Questions if 0 < i <= len(all_tf)]

            for i, mark in enumerate(MCQ_Marks):
                selected_mcq[i]['mark'] = mark

        #--------------------------------------------#
        # Generation and reporting
        tf_questions_all_sets, best_indices, tf_answer_overlap_between_sets = generate_balanced_overlap_sets(selected_tf)
        
        
        mcq_questions_all_sets = generate_shuffled_mcq_sets(selected_mcq)
        tf_answers_all_sets, selected_tf_CA_marks, selected_tf_WA_marks = generate_set_wise_answers_and_marks(
            tf_questions_all_sets, tf_neg_percentage)
        mcq_answers_all_sets, selected_mcq_CA_marks, selected_mcq_WA_marks = generate_set_wise_answers_and_marks(
            mcq_questions_all_sets, mcq_neg_percentage)


        create_correct_answer_excel(correct_answers_template_file,correct_answers_output_file,Question_Name,
                                    tf_answers_all_sets,
                                    selected_tf_CA_marks,
                                    selected_tf_WA_marks,
                                    mcq_answers_all_sets,
                                    selected_mcq_CA_marks,
                                    selected_mcq_WA_marks)


        
        total_tf_marks = sum(selected_tf_CA_marks)
        total_mcq_marks = sum(selected_mcq_CA_marks)
        
        if total_tf_marks + total_mcq_marks != full_mark:
            messages.append("Total mark (mcq+tf) is different from input 'Marks'")
            messages.append("Updated the total marks")
            full_mark = total_tf_marks + total_mcq_marks

        # ✅ Pass the same base_folder for output too (so output paths are consistent)
        output_files_location,generated_pdfs  = create_question_latex_code_and_pdf(
            Department, exam_title, latex_template_file, Question_Name,
            mcq_questions_all_sets, tf_questions_all_sets,
            full_mark, exam_duration, tf_neg_percentage, mcq_neg_percentage,
            base_folder=base_folder
        )

        #--------------------------------------------#
        messages.append("Success!!")
        messages.append(f"{Question_Name}, generated")
        messages.append(f"Total MCQ: {len(selected_mcq)}, T/F: {len(selected_tf)}")

        return JSONResponse({
            "status": "success",
            "messages": messages,
            "exam_questions_folder_location": str(output_files_location),
            "generated_pdfs": generated_pdfs,
            "correct_answer_excel_FN":correct_answer_excel_FN
        })

    except Exception as e:
        print("Error during exam generation:", e)
        traceback.print_exc()
        messages.append(f"❌ Error: {str(e)}")
        return JSONResponse({"status": "error", "messages": messages})



import cv2
import pandas as pd
import numpy as np
widthImg=800
heightImg=800
RectangleScore_dict={}
DisplayLocation={'Roll':1,'TF':2,'QA':3}
DisplayLocation_3img={'Roll':0,'TF':1,'QA':2}
DisplayLocation_4img_large_omr={'SetCode':0,'Roll':1,'TF':2,'QA':3}


Rectangle_Shape_Dict={'QA':(14,9),'TF':(12,4),'Roll':(12,7)}
Rectangle_Shape_Dict_large_omr={'SetCode':(3,3),'QA':(27,17),'TF':(12,12),'Roll':(12,7)}

T='T';F='F';
a, b, c, d, e, f = 'a', 'b', 'c', 'd', 'e', 'f'


Rectangle_no_dict={'QA':0,'Roll':1,'TF':2}
Rectangle_no_dict_large_omr={'QA':0,'TF':1,'Roll':2,'SetCode':3}  #Largest to smallest




def imgPreProcess(img, widthImg, heightImg,rectContours,TakeRectangle,threshold_value=0.5):
    #imgCanny = getImgCanny(img)
    # imgBlank = np.zeros_like(img)
    # contours, hierarchy = cv2.findContours(imgCanny, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
    #rectContours, imgContours = getRectangles_fromImage(img, widthImg, heightImg)

    rectangle_no=Rectangle_no_dict[TakeRectangle]
    selectedContour = getCornerPoints(rectContours[rectangle_no])

    imgWarpColored = getWarpPerspective(img, selectedContour, widthImg, heightImg)
    imgWarpGray = cv2.cvtColor(imgWarpColored, cv2.COLOR_BGR2GRAY)


    # if threshold_value is None:
    #     threshold_value=(int(imgWarpGray.min())+int(imgWarpGray.max()))/2
    # else:
    threshold_value=int(imgWarpGray.min()*threshold_value)+int(imgWarpGray.max()*(1-threshold_value))

    imgThreshold = cv2.threshold(imgWarpGray, threshold_value, 255, cv2.THRESH_BINARY_INV)[1]
    return imgThreshold,imgWarpColored

def getWarpPerspective(img,selectedContour,widthImg,heightImg):
    selectedContour = reorder(selectedContour)
    pt1 = np.float32(selectedContour)
    pt2 = np.float32([[0, 0], [widthImg, 0], [0, heightImg], [widthImg, heightImg]])
    matrix = cv2.getPerspectiveTransform(pt1, pt2)
    # Transform the image into bird-eye-view using the warp transformation matrix
    imgWarpColored = cv2.warpPerspective(img, matrix, (widthImg, heightImg))
    return imgWarpColored

def getRectangles_fromImage(img,widthImg, heightImg):
    img = cv2.resize(img, (widthImg, heightImg))
    imgContours = img.copy()

    imgGray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    imgBlur = cv2.GaussianBlur(imgGray, (5, 5), 1)
    imgCanny = cv2.Canny(imgBlur, 10, 50)  # Canny is edge detector. Here we are able to detect the edge of the rectengles

    ##### FINDING ALL CONTOURS
    contours, hierarchy = cv2.findContours(imgCanny, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
    cv2.drawContours(imgContours, contours, -1, (0, 255, 0), 10)  # Draw all the contours with green line

    #### FIND RECTANGLES
    rectContours = rectContour(contours)  # Find the rectangular contours and sort them according to their area
    return rectContours,imgContours

def rectContour(contours):
    rectCon=[]
    for i in contours:
        area=cv2.contourArea(i)
        #print(area)
        if area>50:
            perimeter=cv2.arcLength(i,True) #Closed perimeter=True
            approx=cv2.approxPolyDP(i,0.02*perimeter,True) #Finds the number of corner points
            #print("Corner Points",len(approx))
            if len(approx)==4:  #taking the rectangles only
                rectCon.append(i)
    rectCon=sorted(rectCon,key=cv2.contourArea,reverse=True)
    return rectCon

def reorder(myPoints):
    #This function takes the corner points of a rectangle as input.
    # Then x+y=minimu, its the origin. If x+y=maximum, it is (w,h)
    #If x-y is minimum, it is (0,h). if x-y is maximu, it is the remaining point (h,0)

    myPoints=myPoints.reshape((4,2))
    myPointsNew_likeVideo=np.zeros((4,1,2),np.int32)
    add=myPoints.sum(1) #

    myPointsNew_likeVideo[0]=myPoints[np.argmin(add)]  #[0,0]
    myPointsNew_likeVideo[3]=myPoints[np.argmax(add)]  #[w,h]
    diff=np.diff(myPoints,axis=1)
    myPointsNew_likeVideo[1]=myPoints[np.argmin(diff)]  #[w,0]
    myPointsNew_likeVideo[2]=myPoints[np.argmax(diff)]  #[0,h]
    return myPointsNew_likeVideo
def getCornerPoints(contour):
    perimeter = cv2.arcLength(contour, True)  # Closed perimeter=True
    approx = cv2.approxPolyDP(contour, 0.02 * perimeter,True)  # Finds the number of corner points
    return approx


def getImgCanny(img):
    imgGray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    imgBlur = cv2.GaussianBlur(imgGray, (5, 5), 1)
    imgCanny = cv2.Canny(imgBlur, 10,50)  # Canny is edge detector. Here we are able to detect the edge of the rectengles
    return imgCanny


def fillFraction(box):
    n_pixel=cv2.countNonZero(box)
    total_pixel=np.shape(box)[0]*np.shape(box)[1]
    return round(n_pixel / total_pixel,4)

def get_Fillups(imgThreshold,n_rows,n_cols,n_skipPixels=30,w=0.5):
    # This function splits a rectangle into individual boxes. Then calculates box fill fraction
    # Note that sometimes box can be significantly larger than the OMR circle. In that case we skip few pixels along
    # the horizontal direction. In other words, we crop the image horizontally
    imgThreshold_modified = cv2.resize(imgThreshold, (n_cols * 100, n_rows * 100))
    rows = np.vsplit(imgThreshold_modified, n_rows)
    ############################
    boxes=[]
    rgbBoxes=[]
    boxFillFractions=[]
    for row in rows:
        cols=np.hsplit(row, n_cols)
        for box in cols:
            grayBox = box.copy()
            rgb_box = cv2.cvtColor(grayBox, cv2.COLOR_GRAY2RGB)

            box=box[:,n_skipPixels:100-n_skipPixels]
            boxes.append(box)
            fill = fillFraction(box)
            boxFillFractions.append(fill)
            #Mark the cropped region with "Blue gray"
            rgb_box[:, 0:n_skipPixels, 1] = 54;rgb_box[:, -n_skipPixels:, 1] = 54
            rgb_box[:, 0:n_skipPixels, 1] = 147;rgb_box[:, -n_skipPixels:, 1] = 147
            rgb_box[:, 0:n_skipPixels, 2] = 179;rgb_box[:, -n_skipPixels:, 2] = 179
            rgbBoxes.append(rgb_box)

    boxes = np.reshape(boxes, (n_rows, n_cols, np.shape(boxes)[1], np.shape(boxes)[2]))
    rgbBoxes=np.reshape(rgbBoxes,(n_rows, n_cols, np.shape(rgbBoxes)[1], np.shape(rgbBoxes)[2],np.shape(rgbBoxes)[3]))
    boxFillFractions = np.reshape(boxFillFractions, (n_rows, n_cols))


    # threshold_row=boxFillFractions[0,:]
    # threshold=w*threshold_row.max()+(1-w)*threshold_row.min()
    threshold=w*boxFillFractions.max()+(1-w)*boxFillFractions.min()

    ind=np.where(boxFillFractions>threshold)
    fillupMatrix=np.zeros(np.shape(boxFillFractions))
    fillupMatrix[ind]=1

    return boxes,rgbBoxes,boxFillFractions,fillupMatrix







answer_loc_dict={'a':1, 'b':2, 'c':4, 'd':5, 'e':7, 'f':8,'T':1,'F':3}

def GenerateCorrectAnswerMatrix_0(Correct_answer_dict,Rectangle_Shape,ignore_nrows=2):

    correct_answer_matrix = np.zeros(Rectangle_Shape)
    for key in Correct_answer_dict:
        correct_answer=Correct_answer_dict[key]  #a, b, c etc.
        column=answer_loc_dict[correct_answer]
        #print(key,column)
        correct_answer_matrix[int(key)+int(ignore_nrows-1), column] = 1
    return correct_answer_matrix

def GenerateCorrectAnswerMatrix(Correct_answer_dict,Rectangle_Shape,ignore_nrows=2,large_omr=False):
    if large_omr:
        answer_loc_dict={'a':1, 'b':2, 'c':3, 'd':5, 'e':6, 'f':7,'T':1,'F':2}
    else:
        answer_loc_dict={'a':1, 'b':2, 'c':4, 'd':5, 'e':7, 'f':8,'T':1,'F':3}
        
    correct_answer_matrix = np.zeros(Rectangle_Shape)
    for key in Correct_answer_dict:
        correct_answer=Correct_answer_dict[key]  #a, b, c etc.
        column=answer_loc_dict[correct_answer]
        #print(key,column)
        correct_answer_matrix[int(key)+int(ignore_nrows-1), column] = 1
    return correct_answer_matrix



def RotateClockwise(*Matrices):
    retList=[]
    for matrix in Matrices:
        matrix=cv2.rotate(matrix, cv2.ROTATE_90_CLOCKWISE)
        retList.append(matrix)
    return retList




def Nullify_DoubleAnswers(fillups,TakeRectangle):
    def ApplyNullification(fillups_innder):
        fillupNullified = fillups_innder.copy()
        for nn in range(0, np.shape(fillups)[0]):
            ind = np.where(fillups[nn,] == 1)
            if len(ind[0])> 1:
                fillupNullified[nn,ind]='Nan'
        return fillupNullified

    fillups[0, :] = 0
    if TakeRectangle=='TF':
        fillupProcessed=ApplyNullification(fillups)
    elif TakeRectangle=='QA':
        if np.shape(fillups)[0]<np.shape(fillups)[1]:
            print("Nullify_DoubleAnswers: Wrong fillup matrix supplied!!! Expecting for 'QA' box...")
            #print(colored("Nullify_DoubleAnswers: Wrong fillup matrix supplied!!! Expecting for 'QA' box..."),'red')
        fillupProcessed = ApplyNullification(fillups)
    else:
        fillupProcessed = ApplyNullification(fillups)
        # fillups[0, :] = 'Nan'
        # fillupProcessed=fillups
    return  fillupProcessed



def computeRoll(fillupMatrix,TakeRectangle):
    fillupMatrix=fillupMatrix[1:,:]  #Skipping 1st row

    if TakeRectangle=='Roll':
        roll=0
        for nn in range(0,7):
            col=fillupMatrix[:,nn]
            ind=np.where(col==1)[0]
            if len(ind)==0:
                print("Roll: No fillup in col ",nn,'\t',col)
                return -1
            elif len(ind)>1:
                print("Roll: Multiple fillup in col ",nn,'\t',col)
                return -1
            else:
                roll=roll+10**(6-nn)*ind[0]
    else:
        roll=-1
    return '{:07.0f}'.format(roll)

def computeRoll_v2(boxFillFractions,TakeRectangle,ignore_nrows=2,w=0.5):
    n_cols=np.shape(boxFillFractions)[1]
    fillupMatrix=np.zeros(np.shape(boxFillFractions))

    threshold=w*boxFillFractions[0,:].max()+(1-w)*boxFillFractions[0,:].min()
    ind=np.where(boxFillFractions[0,:]>threshold)
    fillupMatrix[0,ind]=1

    for nn in range(n_cols):
        ind=np.where(boxFillFractions[ignore_nrows:,nn]==max(boxFillFractions[ignore_nrows:,nn]))[0]
        fillupMatrix[ind+ignore_nrows,nn]=1  #Skipping 1st "ignore_nrows"


    if TakeRectangle=='Roll':
        roll=0
        for nn in range(0,7):
            col=fillupMatrix[ignore_nrows:,nn]
            ind=np.where(col==1)[0]
            if len(ind)==0:
                print("Roll: No fillup in col ",nn,'\t',col)
                return -1
            elif len(ind)>1:
                print("Roll: Multiple fillup in col ",nn,'\t',col)
                return -1
            else:
                roll=roll+10**(6-nn)*ind[0]
    else:
        roll=-1

    return fillupMatrix,'{:07.0f}'.format(roll)

# def computeRoll_v2_long_omr(boxFillFractions,TakeRectangle,ignore_nrows=2,w=0.5):
    # n_cols=np.shape(boxFillFractions)[1]
    # fillupMatrix=np.zeros(np.shape(boxFillFractions))

    # threshold=w*boxFillFractions[0,:].max()+(1-w)*boxFillFractions[0,:].min()
    # ind=np.where(boxFillFractions[0,:]>threshold)
    # fillupMatrix[0,ind]=1

    # for nn in range(n_cols):
        # ind=np.where(boxFillFractions[ignore_nrows:,nn]==max(boxFillFractions[ignore_nrows:,nn]))[0]
        # fillupMatrix[ind+ignore_nrows,nn]=1  #Skipping 1st "ignore_nrows"


    # if TakeRectangle=='Roll':
        # roll=0
        # for nn in range(0,7):
            # col=fillupMatrix[ignore_nrows:,nn]
            # ind=np.where(col==1)[0]
            # if len(ind)==0:
                # print("Roll: No fillup in col ",nn,'\t',col)
                # return -1
            # elif len(ind)>1:
                # print("Roll: Multiple fillup in col ",nn,'\t',col)
                # return -1
            # else:
                # roll=roll+10**(6-nn)*ind[0]
    # else:
        # roll=-1

    # return fillupMatrix,'{:07.0f}'.format(roll)


import itertools

def showRoll(imgWarpColored, fillupMatrix_doubleAns_nullified):
    imgAnswers = imgWarpColored.copy()

    n_rows = np.shape(fillupMatrix_doubleAns_nullified)[0]
    n_cols = np.shape(fillupMatrix_doubleAns_nullified)[1]


    secW = int(imgAnswers.shape[1] / n_cols)
    secH = int(imgAnswers.shape[0] / n_rows)
    thickness = 12

    for row, col in itertools.product(range(1,n_rows), range(n_cols)):
        if np.isnan(fillupMatrix_doubleAns_nullified[row,]).any():  # If row contains a 'nan' (multiple answer),continue
            cX=int(secW / 2 + 0 * secW)
            cY = int(secH / 2 + row * secH)
            cv2.rectangle(imgAnswers, (int(cX - secW / 2), int(cY+ secH / 2)),
                          (int(cX+n_rows* secW + secW / 2), int(cY - secH / 2)), (94, 48, 201), thickness)
            continue

        Entry_Student_answerMatrix = fillupMatrix_doubleAns_nullified[row, col]

        if Entry_Student_answerMatrix:   #Mark the answer only if student answered it
            cX_student = int(secW / 2 + col * secW)
            cY_student = int(secH / 2 + row * secH)

            cv2.rectangle(imgAnswers, (int(cX_student-secW / 2),int(cY_student+secH/2)), (int(cX_student+secW / 2),int(cY_student-secH/2)), (255, 255, 0), thickness)
    return imgAnswers

def show_SetCode(imgWarpColored, fillupMatrix_doubleAns_nullified):
    imgAnswers = imgWarpColored.copy()

    n_rows = np.shape(fillupMatrix_doubleAns_nullified)[0]
    n_cols = np.shape(fillupMatrix_doubleAns_nullified)[1]


    secW = int(imgAnswers.shape[1] / n_cols)
    secH = int(imgAnswers.shape[0] / n_rows)
    thickness = 35

    for row, col in itertools.product(range(1,n_rows), range(n_cols)):
        if np.isnan(fillupMatrix_doubleAns_nullified[row,]).any():  # If row contains a 'nan' (multiple answer),continue
            cX=int(secW / 2 + 0 * secW)
            cY = int(secH / 2 + row * secH)
            cv2.rectangle(imgAnswers, (int(cX - secW / 2), int(cY+ secH / 2)),
                          (int(cX+n_rows* secW + secW / 2), int(cY - secH / 2)), (94, 48, 201), thickness)
            continue

        Entry_Student_answerMatrix = fillupMatrix_doubleAns_nullified[row, col]

        if Entry_Student_answerMatrix:   #Mark the answer only if student answered it
            cX_student = int(secW / 2 + col * secW)
            cY_student = int(secH / 2 + row * secH)

            cv2.rectangle(imgAnswers, (int(cX_student-secW / 2),int(cY_student+secH/2)), (int(cX_student+secW / 2),int(cY_student-secH/2)), (255, 255, 0), thickness)
    return imgAnswers





def Convert_integer_mark_into_dictionary(Mark_correctAns_dict,Mark_wrongAns_dict,dictLen):
    if type(Mark_correctAns_dict) is not dict:
        tmp_Mark_correctAnswer = {}
        tmr_Mark_wrongAnswer = {}
        for row in range(0, dictLen):
            tmp_Mark_correctAnswer[row] = Mark_correctAns_dict
            tmr_Mark_wrongAnswer[row] = Mark_wrongAns_dict
        Mark_correctAns_dict = tmp_Mark_correctAnswer
        Mark_wrongAns_dict = tmr_Mark_wrongAnswer
    return Mark_correctAns_dict,Mark_wrongAns_dict


def GenerateScore(TakeRectangle,Mark_dict,correct_answer_matrix,fillupMatrix_doubleAns_nullified,imgWarpColored,ignore_nrows=2):
    def _GenerateScore(correct_answer_matrix,fillupMatrix_doubleAns_nullified,Mark_correctAnswer_dict,Mark_wrongAnswer_dict,imgWarpColored,ignore_nrows):
        print("_GenerateScore is called")
        n_rows = np.shape(correct_answer_matrix)[0]
        n_cols = np.shape(correct_answer_matrix)[1]
        
        
        text_x_offset_dict = {'TF': 2, 'QA': 4}
        mark_x_offset_dict = {'TF': 0.6, 'QA': 3}
        text_x_offset = text_x_offset_dict[TakeRectangle]
        mark_x_offset_dict = mark_x_offset_dict[TakeRectangle]
        imgAnswers = imgWarpColored.copy()
        secW = int(imgAnswers.shape[1] / n_cols)
        secH = int(imgAnswers.shape[0] / n_rows)
        angle = 0;
        startAngle = 0;
        endAngle = 360;
        axesLength = (int(secW / 2.5), int(secH / 2));
        thickness = 8

        score_positive = 0
        score_negative = 0
        text_thickness=4
        
        print("Starting row loop...")
        for row in range(ignore_nrows, n_rows):
            # if Mark_correctAnswer is None or Mark_wrongAnswer is None:  #If the mark is not given
            #     continue
            
            print("row= ",row)
            Row_studentAnswerMatrix = fillupMatrix_doubleAns_nullified[row, :]
            Row_CorrectAnswerMatrix = correct_answer_matrix[row, :]
            Dict_key = row - (ignore_nrows - 1)
            print(f"Line 418, Row_studentAnswerMatrix={Row_studentAnswerMatrix},Row_CorrectAnswerMatrix{Row_CorrectAnswerMatrix} ")
            if not np.any(Row_studentAnswerMatrix):  #Student did not answer this question.
                continue
            if np.isnan(Row_studentAnswerMatrix).any():  # Multiple answering nullifies the answer.
                cX = int(secW / 2 + 0 * secW)
                cY = int(secH / 2 + row * secH)
                cv2.rectangle(imgAnswers, (int(cX - secW / 2), int(cY + secH / 2)),(int(cX + n_rows * secW + secW / 2), int(cY - secH / 2)), (94, 48, 201), thickness)
                text_coordinate = (int(secW * text_x_offset), cY)
                cv2.putText(imgAnswers, "Null", text_coordinate, cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 0, 255), text_thickness, cv2.LINE_AA)
                continue

            print("Line 429")
            studentAnsIndex = np.where(Row_studentAnswerMatrix == 1)[0]
            correctAnsIndex = np.where(Row_CorrectAnswerMatrix == 1)[0]
            print("Line 432")
            
            if studentAnsIndex.size == 0:
                print("Student did not answer the question...")
                continue   # or handle however you want

            
            cX_student = int(secW / 2 + studentAnsIndex[0] * secW)
            cY_student = int(secH / 2 + row * secH)
            student_fillup_coordinate = (cX_student, cY_student);
            Mark_text_coordinate = (int(secW * mark_x_offset_dict), cY_student)
            #        Mark_text_coordinate=(int(secW / 2), cY_student)
            text_coordinate = (int(secW * text_x_offset), cY_student)

            print("Line 440   Text coordinate computed")
#            print(TakeRectangle,"Mark=",Mark_correctAnswer_dict)
            print("[442] Row_CorrectAnswerMatrix",Row_CorrectAnswerMatrix)
            print("[443 Row_studentAnswerMatrix]",Row_studentAnswerMatrix)
            if not np.any(Row_CorrectAnswerMatrix):  # If there is no "correct answer" of a given question, though student answered it.
                                                     # In that case, assume this question does not exist, and continue.
                print("Line 445")
                cv2.rectangle(imgAnswers, (int(cX_student-secW / 2),int(cY_student+secH/2)), (int(cX_student+secW / 2),int(cY_student-secH/2)), (255, 255, 0), thickness)
                print("Line 447")
                cv2.putText(imgAnswers,"Ans. Not Provided",text_coordinate,cv2.FONT_HERSHEY_SIMPLEX,1,(0,0,255),4,cv2.LINE_AA)
                continue

            elif studentAnsIndex == correctAnsIndex:    #Student filled the correct circle
                print("   Line 453")
                cv2.ellipse(imgAnswers, student_fillup_coordinate, axesLength, angle, startAngle, endAngle, (0, 255, 0),thickness)
                if Mark_correctAnswer_dict.get(Dict_key) is not None and not np.isnan(Mark_correctAnswer_dict.get(Dict_key)):
                    Mark_correctAnswer = Mark_correctAnswer_dict.get(Dict_key)
                    score_positive = score_positive + Mark_correctAnswer
                    cv2.putText(imgAnswers,str(Mark_correctAnswer),Mark_text_coordinate,cv2.FONT_HERSHEY_SIMPLEX,1,(0,0,255),4,cv2.LINE_AA)
                else:
                    cv2.putText(imgAnswers, "CA Mark Not provided", text_coordinate, cv2.FONT_HERSHEY_SIMPLEX, 1,(255, 0, 255), 4,cv2.LINE_AA)

                print("   Line 462 passed")
            else: #student filled a wrong circle
                print("   Line 464")
                cv2.ellipse(imgAnswers, student_fillup_coordinate, axesLength, angle, startAngle, endAngle, (0, 0, 255),thickness)
                print("   Line 466")
                col = np.where(correct_answer_matrix[row,] == 1)[0][0]
                print(f"   Line 468, col={col}")
                cX_correct = int(secW / 2 + col * secW); cY_correct = int(secH / 2 + row * secH)
                print("   Line 470")
                cv2.ellipse(imgAnswers, (cX_correct, cY_correct), axesLength, angle, startAngle, endAngle, (255, 255, 0), thickness)

                if Mark_wrongAnswer_dict.get(Dict_key) is not None and not np.isnan(Mark_wrongAnswer_dict.get(Dict_key)):
                    Mark_wrongAnswer = Mark_wrongAnswer_dict.get(Dict_key)
                    score_negative = score_negative + Mark_wrongAnswer
                    cv2.putText(imgAnswers,str(Mark_wrongAnswer),Mark_text_coordinate,cv2.FONT_HERSHEY_SIMPLEX,1,(0,0,255),4,cv2.LINE_AA)
                else:
                    cv2.putText(imgAnswers, "WA Mark Not provided", text_coordinate, cv2.FONT_HERSHEY_SIMPLEX, 1,(255, 0, 255), 4, cv2.LINE_AA)

        return score_positive,score_negative,imgAnswers
    ##################################  showAnswers  works from Here ###############################################

    print("GenerateScore is called in the utils")
    Mark_correctAnswer_dict,Mark_wrongAnswer_dict=Mark_dict[TakeRectangle]
    dictLen=np.shape(correct_answer_matrix)[0]
    Mark_correctAnswer_dict, Mark_wrongAnswer_dict = Convert_integer_mark_into_dictionary(Mark_correctAnswer_dict,Mark_wrongAnswer_dict,dictLen)

    print(f"[Generate Score:] Mark_correct_Answer_dict is created. They are {Mark_correctAnswer_dict} and {Mark_wrongAnswer_dict}")
    score_positive,score_negative,imgAnswers=_GenerateScore(correct_answer_matrix,fillupMatrix_doubleAns_nullified,Mark_correctAnswer_dict, Mark_wrongAnswer_dict,imgWarpColored,ignore_nrows)

    return score_positive,score_negative,imgAnswers












def Conv_FirstRow_into_Binary(a):
    b = 0
    for i in range(len(a)):
        b=b+10**i*a[len(a)-i-1]
    return int(b)

def ReadCorrectAnswer(xlsxFN, nrows=12, usecols='A:Q', Print=False):
    print("\n\t\tReading "+str(xlsxFN)+'...')
    AllAnswer_df = pd.read_excel(xlsxFN, nrows=nrows, usecols=usecols)

    TF_df = AllAnswer_df[
        ['True/False no', 'T/F Set 1', 'T/F Set 2', 'T/F Set 3', 'Correct T/F mark', 'Wrong T/F mark']].copy()
    QA_df = AllAnswer_df[
        ['Question no', 'Q/A Set 1', 'Q/A Set 2', 'Q/A Set 3', 'Correct Q/A mark', 'Wrong Q/A mark']].copy()

    Course = list(AllAnswer_df['Course'])[0]

    binaryCode=AllAnswer_df['Binary code'].dropna()
    binaryCode=list(binaryCode.astype(int))
    equivalentSetCode=AllAnswer_df['Equivalent Set Code'].dropna()
    equivalentSetCode =list(equivalentSetCode.astype(int))
    SetCode = pd.Series(equivalentSetCode,index=binaryCode)
    GetSetCode_from_Binary=SetCode.to_dict()


#    TF_df = TF_df.dropna(subset=['T/F Set 1', 'T/F Set 2', 'T/F Set 3', 'Correct T/F mark', 'Wrong T/F mark'])
#    QA_df = QA_df.dropna(subset=['Q/A Set 1', 'Q/A Set 2', 'Q/A Set 3', 'Correct Q/A mark', 'Wrong Q/A mark'])

    if Print:
        print('\n\t Verify the content of '+str(xlsxFN)+':\n')
        print('\t\tCourse= ',Course)
        for key in GetSetCode_from_Binary:
            print("\t\tBinary:", key, '\t\tSet Code:', GetSetCode_from_Binary[key])
        print('\n')
        print(TF_df.to_string(index=False), '\n', QA_df.to_string(index=False))

    TF_question_no = TF_df['True/False no']
    TF_set1 = TF_df['T/F Set 1']
    TF_set2 = TF_df['T/F Set 2']
    TF_set3 = TF_df['T/F Set 3']
    C_TF_Mark_Arr = TF_df['Correct T/F mark']
    W_TF_Mark_Arr = TF_df['Wrong T/F mark']

    QA_question_no = QA_df['Question no']
    QA_set1 = QA_df['Q/A Set 1']
    QA_set2 = QA_df['Q/A Set 2']
    QA_set3 = QA_df['Q/A Set 3']
    C_QA_Mark_Arr = QA_df['Correct Q/A mark']
    W_QA_Mark_Arr = QA_df['Wrong Q/A mark']


    
    TF_ind_set1={}
    TF_ind_set2 = {}
    TF_ind_set3 = {}
    Mark_correctAnswer_TF={}
    Mark_wrongAnswer_TF={}


    for i in range(len(TF_question_no)):
        if isinstance(TF_set1[i], str):
            TF_ind_set1[TF_question_no[i]] = TF_set1[i].strip()
        if isinstance(TF_set2[i], str):
            TF_ind_set2[TF_question_no[i]] = TF_set2[i].strip()
        if isinstance(TF_set3[i], str):
            TF_ind_set3[TF_question_no[i]] = TF_set3[i].strip()
        Mark_correctAnswer_TF[TF_question_no[i]] = C_TF_Mark_Arr[i]
        Mark_wrongAnswer_TF[TF_question_no[i]] = W_TF_Mark_Arr[i]

    QA_ind_set1={}
    QA_ind_set2 = {}
    QA_ind_set3 = {}
    Mark_CorrectAnswer_QA={}
    Mark_wrongAnswer_QA={}

    for i in range(len(QA_question_no)):
        if isinstance(QA_set1[i], str):
            QA_ind_set1[QA_question_no[i]] = QA_set1[i].strip()
        if isinstance(QA_set2[i], str):
            QA_ind_set2[QA_question_no[i]] = QA_set2[i].strip()
        if isinstance(QA_set3[i], str):
            QA_ind_set3[QA_question_no[i]] = QA_set3[i].strip()
        Mark_CorrectAnswer_QA[QA_question_no[i]] = C_QA_Mark_Arr[i]
        Mark_wrongAnswer_QA[QA_question_no[i]] = W_QA_Mark_Arr[i]

    return Course,GetSetCode_from_Binary,TF_ind_set1, TF_ind_set2, TF_ind_set3, QA_ind_set1, QA_ind_set2, QA_ind_set3, Mark_correctAnswer_TF, Mark_wrongAnswer_TF, Mark_CorrectAnswer_QA, Mark_wrongAnswer_QA


def ReadCorrectAnswer_dict_from_webread_df(AllAnswer_df, Print=False):

    TF_df = AllAnswer_df[
        ['True/False no', 'T/F Set 1', 'T/F Set 2', 'T/F Set 3', 'Correct T/F mark', 'Wrong T/F mark']].copy()
    QA_df = AllAnswer_df[
        ['Question no', 'Q/A Set 1', 'Q/A Set 2', 'Q/A Set 3', 'Correct Q/A mark', 'Wrong Q/A mark']].copy()

    Course = list(AllAnswer_df['Course'])[0]

    binaryCode=AllAnswer_df['Binary code'].dropna()
    binaryCode=list(binaryCode.astype(int))
    equivalentSetCode=AllAnswer_df['Equivalent Set Code'].dropna()
    equivalentSetCode =list(equivalentSetCode.astype(int))
    SetCode = pd.Series(equivalentSetCode,index=binaryCode)
    GetSetCode_from_Binary=SetCode.to_dict()


#    TF_df = TF_df.dropna(subset=['T/F Set 1', 'T/F Set 2', 'T/F Set 3', 'Correct T/F mark', 'Wrong T/F mark'])
#    QA_df = QA_df.dropna(subset=['Q/A Set 1', 'Q/A Set 2', 'Q/A Set 3', 'Correct Q/A mark', 'Wrong Q/A mark'])

    if Print:
        print('\n\t Verify the content of '+str(xlsxFN)+':\n')
        print('\t\tCourse= ',Course)
        for key in GetSetCode_from_Binary:
            print("\t\tBinary:", key, '\t\tSet Code:', GetSetCode_from_Binary[key])
        print('\n')
        print(TF_df.to_string(index=False), '\n', QA_df.to_string(index=False))

    TF_question_no = TF_df['True/False no']
    TF_set1 = TF_df['T/F Set 1']
    TF_set2 = TF_df['T/F Set 2']
    TF_set3 = TF_df['T/F Set 3']
    C_TF_Mark_Arr = TF_df['Correct T/F mark']
    W_TF_Mark_Arr = TF_df['Wrong T/F mark']

    QA_question_no = QA_df['Question no']
    QA_set1 = QA_df['Q/A Set 1']
    QA_set2 = QA_df['Q/A Set 2']
    QA_set3 = QA_df['Q/A Set 3']
    C_QA_Mark_Arr = QA_df['Correct Q/A mark']
    W_QA_Mark_Arr = QA_df['Wrong Q/A mark']


    
    TF_ind_set1={}
    TF_ind_set2 = {}
    TF_ind_set3 = {}
    Mark_correctAnswer_TF={}
    Mark_wrongAnswer_TF={}


    for i in range(len(TF_question_no)):
        if isinstance(TF_set1[i], str):
            TF_ind_set1[TF_question_no[i]] = TF_set1[i].strip()
        if isinstance(TF_set2[i], str):
            TF_ind_set2[TF_question_no[i]] = TF_set2[i].strip()
        if isinstance(TF_set3[i], str):
            TF_ind_set3[TF_question_no[i]] = TF_set3[i].strip()
        Mark_correctAnswer_TF[TF_question_no[i]] = C_TF_Mark_Arr[i]
        Mark_wrongAnswer_TF[TF_question_no[i]] = W_TF_Mark_Arr[i]

    QA_ind_set1={}
    QA_ind_set2 = {}
    QA_ind_set3 = {}
    Mark_CorrectAnswer_QA={}
    Mark_wrongAnswer_QA={}

    for i in range(len(QA_question_no)):
        if isinstance(QA_set1[i], str):
            QA_ind_set1[QA_question_no[i]] = QA_set1[i].strip()
        if isinstance(QA_set2[i], str):
            QA_ind_set2[QA_question_no[i]] = QA_set2[i].strip()
        if isinstance(QA_set3[i], str):
            QA_ind_set3[QA_question_no[i]] = QA_set3[i].strip()
        Mark_CorrectAnswer_QA[QA_question_no[i]] = C_QA_Mark_Arr[i]
        Mark_wrongAnswer_QA[QA_question_no[i]] = W_QA_Mark_Arr[i]

    return Course,GetSetCode_from_Binary,TF_ind_set1, TF_ind_set2, TF_ind_set3, QA_ind_set1, QA_ind_set2, QA_ind_set3, Mark_correctAnswer_TF, Mark_wrongAnswer_TF, Mark_CorrectAnswer_QA, Mark_wrongAnswer_QA




def stackImages(imgArray,scale):
    rows = len(imgArray)
    cols = len(imgArray[0])

    rowsAvailable = isinstance(imgArray[0], (list,np.ndarray))
    width = imgArray[0][0].shape[1]
    height = imgArray[0][0].shape[0]

    if rowsAvailable:
        for x in range ( 0, rows):
            for y in range(0, cols):
                if imgArray[x][y].shape[:2] == imgArray[0][0].shape [:2]:
                    imgArray[x][y] = cv2.resize(imgArray[x][y], (0, 0), None, scale, scale)
                else:
                    imgArray[x][y] = cv2.resize(imgArray[x][y], (imgArray[0][0].shape[1], imgArray[0][0].shape[0]), None, scale, scale)
                if len(imgArray[x][y].shape) == 2: imgArray[x][y]= cv2.cvtColor( imgArray[x][y], cv2.COLOR_GRAY2BGR)
        imageBlank = np.zeros((height, width, 3), np.uint8)
        hor = [imageBlank]*rows
        hor_con = [imageBlank]*rows
        for x in range(0, rows):
            hor[x] = np.hstack(imgArray[x])
        ver = np.vstack(hor)

    else:
        for x in range(0, rows):
            if imgArray[x].shape[:2] == imgArray[0].shape[:2]:
                imgArray[x] = cv2.resize(imgArray[x], (0, 0), None, scale, scale)
            else:
                imgArray[x] = cv2.resize(imgArray[x], (imgArray[0].shape[1], imgArray[0].shape[0]), None,scale, scale)
            if len(imgArray[x].shape) == 2: imgArray[x] = cv2.cvtColor(imgArray[x], cv2.COLOR_GRAY2BGR)
        hor= np.hstack(imgArray)
        ver = hor
    return ver


###################################################################################################
####################   Long OMR related Functions #################################################
###################################################################################################


def GenerateScore_long_omr(TakeRectangle,Mark_dict,correct_answer_matrix,fillupMatrix_doubleAns_nullified,imgWarpColored,ignore_nrows=2):
    def compute_rectangle_coord(Row_studentAnswerMatrix,col_offset,row_offset,row_num,ignore_nrows,secW,secH,Ans_not_Provided=False):
        if Ans_not_Provided:
            idx_nan=[0]
        else:              
            idx_nan = np.where(np.isnan(Row_studentAnswerMatrix))[0]
        
        print("row_num=",row_num,idx_nan)        
        cX_start = int((idx_nan[0] + col_offset) * secW)
        cy_start = int(-secH + (row_num + ignore_nrows - row_offset) * secH)
        cX_end   = int((idx_nan[-1] + col_offset) * secW + secW)
        cY_end   = int((row_num + ignore_nrows - row_offset) * secH)
        return cX_start, cy_start, cX_end, cY_end        
    
    def compute_marker_coordinates(TakeRectangle,studentAnsIndex, correctAnsIndex,col_offset,row_num, ignore_nrows, row_offset,secW, secH):
        # Student bubble center
        cX_student = int(secW/2 + (studentAnsIndex + col_offset) * secW)
        cY_student = int((row_num + ignore_nrows - row_offset) * secH - secH/2)

        # Correct answer bubble center
        print("correctAnsIndex",correctAnsIndex,"col_offset=",col_offset)
        cX_correct_ans = int(secW/2 + (correctAnsIndex + col_offset) * secW)
        cY_correct_ans = cY_student  # same row

        # Mark location depends on block
        TF_mark_location_lookup_value = {0: 3, 4: 7, 8: 11, 9:4,25:13}
        QA_mark_location_lookup_value ={0: 4, 9:13}
        if TakeRectangle=='TF':
            location_lookup_value=TF_mark_location_lookup_value
        elif TakeRectangle=='QA':
            location_lookup_value=QA_mark_location_lookup_value
            

        mark_location_multiplier = location_lookup_value[col_offset]
        
        cX_mark = int(secW * mark_location_multiplier)
        cY_mark = cY_student+int(secW/5)

        # Final coordinate tuples
        student_fillup_coordinate = (cX_student, cY_student)
        CA_circle_coordinate = (cX_correct_ans, cY_correct_ans)
        text_coordinate = (cX_student, cY_student)
        Mark_text_coordinate = (cX_mark, cY_mark)
        return student_fillup_coordinate, CA_circle_coordinate, text_coordinate, Mark_text_coordinate

    def _GenerateScore(TakeRectangle,correct_answer_matrix,fillupMatrix_doubleAns_nullified,Mark_correctAnswer_dict,Mark_wrongAnswer_dict,imgWarpColored,ignore_nrows):        
        
        print("[SubFunc] _GenerateScore is called")
        if TakeRectangle=='TF':
            n_rows=32
            n_cols=4
            n_rows_img=12
            n_cols_img=12            
        elif TakeRectangle=='QA':
            n_rows=52
            n_cols=9
            n_rows_img=27
            n_cols_img=17

        imgAnswers = imgWarpColored.copy()
        secW = imgAnswers.shape[1] / n_cols_img
        secH = imgAnswers.shape[0] / n_rows_img
        angle = 0;
        startAngle = 0;
        endAngle = 360;
        axesLength = (int(secW / 2.5), int(secH / 2));
        
        
        score_positive = 0
        score_negative = 0
        print("[_GenerateScore] Initial setup success")
#        for row in range(ignore_nrows, n_rows):
        for row_num, row in enumerate(range(ignore_nrows, n_rows), start=1):
            
            Row_studentAnswerMatrix = fillupMatrix_doubleAns_nullified[row, :]
            Row_CorrectAnswerMatrix = correct_answer_matrix[row, :]
            correctAnsIndex = np.where(Row_CorrectAnswerMatrix == 1)[0]
            #print("row_num",row_num,"Row_studentAnswerMatrix=",Row_studentAnswerMatrix,"Row_CorrectAnswerMatrix",Row_CorrectAnswerMatrix)
            
            print(f" for Row {row_num} rows are successfuly selected")
            if not np.any(Row_studentAnswerMatrix):  #Student did not answer this question.
                continue
                     
            studentAnsIndex = np.where(Row_studentAnswerMatrix == 1)[0]     
            
            if TakeRectangle == 'TF':
                thickness = 8
                text_thickness=4
                text_font_size=1
                if row_num > 20:
                    col_offset = 8
                    row_offset = 20
                elif row_num > 10:
                    col_offset = 4
                    row_offset = 10
                else:
                    col_offset = 0
                    row_offset = 0
            if TakeRectangle == 'QA':
                thickness = 4
                text_thickness=2
                text_font_size=0.8
                if row_num > 25:
                    col_offset = 9
                    row_offset = 25
                else:
                    col_offset = 0
                    row_offset = 0

            if not np.any(Row_CorrectAnswerMatrix):  # If there is no "correct answer" of a given question, though student answered it.
                print("row_num",row_num," Answer not provided")
                cX_start, cy_start, cX_end, cY_end=compute_rectangle_coord(Row_studentAnswerMatrix,col_offset,row_offset,row_num,ignore_nrows,secW,secH,Ans_not_Provided=True)
                cv2.rectangle(imgAnswers, (cX_start, cy_start), (cX_end, cY_end), (0, 0, 255), thickness)
                continue


            ############ Identify where to mark(put circles) the script ######################
            if np.isnan(Row_studentAnswerMatrix).any():   #Student filled up Multiple circles
                cX_start, cy_start, cX_end, cY_end=compute_rectangle_coord(Row_studentAnswerMatrix,col_offset,row_offset,row_num,ignore_nrows,secW,secH)
                cv2.rectangle(imgAnswers, (cX_start, cy_start), (cX_end, cY_end), (94, 48, 201), thickness)
                continue
            else:  #Student filled up one circle
                student_fillup_coordinate, CA_circle_coordinate, text_coordinate, Mark_text_coordinate=compute_marker_coordinates(TakeRectangle,studentAnsIndex, correctAnsIndex,col_offset,row_num, ignore_nrows, row_offset,secW, secH)
            ############ Identify where to mark(put circles) the script ######################
          
            
            
           ######################################## Scoring of the script #############################
           
            Dict_key = row - (ignore_nrows - 1)
            if studentAnsIndex == correctAnsIndex:    #Student filled the correct circle
                cv2.ellipse(imgAnswers, student_fillup_coordinate, axesLength, angle, startAngle, endAngle, (0, 255, 0),thickness)                                       
                if Mark_correctAnswer_dict.get(Dict_key) is not None and not np.isnan(Mark_correctAnswer_dict.get(Dict_key)):
                    Mark_correctAnswer = Mark_correctAnswer_dict.get(Dict_key)
                    score_positive = score_positive + Mark_correctAnswer
                    cv2.putText(imgAnswers,str(Mark_correctAnswer),Mark_text_coordinate,cv2.FONT_HERSHEY_SIMPLEX,text_font_size,(0,0,255),text_thickness,cv2.LINE_AA)
                else:
                    cv2.putText(imgAnswers, "CA Mark Not provided", text_coordinate, cv2.FONT_HERSHEY_SIMPLEX, text_font_size,(255, 0, 255), text_thickness,cv2.LINE_AA)


            else: #student filled a wrong circle
                cv2.ellipse(imgAnswers, student_fillup_coordinate, axesLength, angle, startAngle, endAngle, (0, 0, 255),thickness)
                cv2.ellipse(imgAnswers, CA_circle_coordinate, axesLength, angle, startAngle, endAngle, (255, 255, 0), thickness)

                if Mark_wrongAnswer_dict.get(Dict_key) is not None and not np.isnan(Mark_wrongAnswer_dict.get(Dict_key)):
                    Mark_wrongAnswer = Mark_wrongAnswer_dict.get(Dict_key)
                    score_negative = score_negative + Mark_wrongAnswer
                    cv2.putText(imgAnswers,str(Mark_wrongAnswer),Mark_text_coordinate,cv2.FONT_HERSHEY_SIMPLEX,text_font_size,(0,0,255),text_thickness,cv2.LINE_AA)
                else:
                    cv2.putText(imgAnswers, "WA Mark Not provided", text_coordinate, cv2.FONT_HERSHEY_SIMPLEX, text_font_size,(255, 0, 255), text_thickness, cv2.LINE_AA)
           ######################################## Scoring of the script #############################
           
        return score_positive,score_negative,imgAnswers
    ##################################  showAnswers  works from Here ###############################################
    print("GenerateScore_long_omr is called")

    Mark_correctAnswer_dict,Mark_wrongAnswer_dict=Mark_dict[TakeRectangle]
    dictLen=np.shape(correct_answer_matrix)[0]
    Mark_correctAnswer_dict, Mark_wrongAnswer_dict = Convert_integer_mark_into_dictionary(Mark_correctAnswer_dict,Mark_wrongAnswer_dict,dictLen)

    print("Calling _GenerateScore,", TakeRectangle)
    score_positive,score_negative,imgAnswers=_GenerateScore(TakeRectangle,correct_answer_matrix,fillupMatrix_doubleAns_nullified,Mark_correctAnswer_dict, Mark_wrongAnswer_dict,imgWarpColored,ignore_nrows)

    return score_positive,score_negative,imgAnswers
    
    
def get_sec_code_long_omr(fillupMatrix):
    set_code_large_omr=None
    if np.sum(fillupMatrix)==1:
        if fillupMatrix[1,2]==1:
            set_code_large_omr=1
        elif fillupMatrix[1,1]==1:
            set_code_large_omr=2
        elif fillupMatrix[1,0]==1:
            set_code_large_omr=3

    return set_code_large_omr

def reshape_fillupMatrix(fillupMatrix,TakeRectangle):
    def reshape_TF_long_omr(fillupMatrix,validate=False):
        top    = fillupMatrix[:12,  0:4]   # rows 0–9,   cols 0–3
        middle = fillupMatrix[2:12,  4:8]   # rows 0–9,   cols 4–7
        bottom = fillupMatrix[2:12,  8:13]  # rows 0–9,   cols 8–12
        
        fillupMatrix_resized =np.vstack([top, middle, bottom])
        if validate:
            for i, row in enumerate(fillupMatrix_resized):
                print(f"Question {i-1}: {row}")
        return fillupMatrix_resized
    def reshape_QA_long_omr(fillupMatrix,validate=False):
        top    = fillupMatrix[:28,  0:8]   # rows 0–27,   cols 0–8
        bottom = fillupMatrix[2:28,  9:18]  # rows 0–9,   cols 9–17
        
        fillupMatrix_resized =np.vstack([top, bottom])
        if validate:
            for i, row in enumerate(fillupMatrix_resized):
                print(f"Question {i-1}: {row}")
        return fillupMatrix_resized
    
    if TakeRectangle=='TF':
        fillupMatrix_resized=reshape_TF_long_omr(fillupMatrix,validate=False)
    elif TakeRectangle=='QA':
        fillupMatrix_resized=reshape_QA_long_omr(fillupMatrix,validate=False)
    return fillupMatrix_resized


def imgPreProcess_large_omr(img, widthImg, heightImg,rectContours,TakeRectangle,threshold_value=0.5):
    rectangle_no=Rectangle_no_dict_large_omr[TakeRectangle]
    selectedContour = getCornerPoints(rectContours[rectangle_no])

    imgWarpColored = getWarpPerspective(img, selectedContour, widthImg, heightImg)
    imgWarpGray = cv2.cvtColor(imgWarpColored, cv2.COLOR_BGR2GRAY)

    threshold_value=int(imgWarpGray.min()*threshold_value)+int(imgWarpGray.max()*(1-threshold_value))

    imgThreshold = cv2.threshold(imgWarpGray, threshold_value, 255, cv2.THRESH_BINARY_INV)[1]
    return imgThreshold,imgWarpColored


###################################################################################################
####################   Long OMR related Functions #################################################
###################################################################################################


def order_points(pts):
    """
    Order points as: top-left, top-right, bottom-right, bottom-left
    pts may come in shape (4,1,2) or (4,2)
    """
    pts = np.asarray(pts)
    pts = pts.reshape(4, 2).astype("float32")

    rect = np.zeros((4, 2), dtype="float32")
    s = pts.sum(axis=1)
    rect[0] = pts[np.argmin(s)]      # top-left: smallest sum
    rect[2] = pts[np.argmax(s)]      # bottom-right: largest sum

    diff = np.diff(pts, axis=1).reshape(4)
    rect[1] = pts[np.argmin(diff)]   # top-right: smallest diff
    rect[3] = pts[np.argmax(diff)]   # bottom-left: largest diff

    return rect

def putWarpBack(TakeRectangle, rectContours, orig_img, processed_img):
    """
    Places processed_img back into orig_img at the quadrilateral defined by
    rectContours[ rectangle_no ] where rectangle_no = Rectangle_no_dict_large_omr[TakeRectangle]

    - processed_img may be any size (e.g. 800x800); it will be resized to fit the target area.
    - Returns a new image (copy of orig_img with the processed region placed back).
    """
    # --- get contour points and order them ---
    rectangle_no = Rectangle_no_dict_large_omr[TakeRectangle]
    contour_points = getCornerPoints(rectContours[rectangle_no])   # whatever shape your helper returns
    dst_pts = order_points(contour_points)  # TL, TR, BR, BL

    # --- compute target width and height from contour geometry ---
    (tl, tr, br, bl) = dst_pts

    widthA = np.linalg.norm(br - bl)   # bottom width
    widthB = np.linalg.norm(tr - tl)   # top width
    target_w = int(round(max(widthA, widthB)))

    heightA = np.linalg.norm(tr - br)  # right height
    heightB = np.linalg.norm(tl - bl)  # left height
    target_h = int(round(max(heightA, heightB)))

    # guard against zero sizes
    if target_w < 1: 
        target_w = 1
    if target_h < 1:
        target_h = 1

    # --- resize processed_img to target size (this preserves the content by stretching/squeezing) ---
    resized = cv2.resize(processed_img, (target_w, target_h), interpolation=cv2.INTER_AREA)

    # --- source points for resized image: TL, TR, BR, BL ---
    src_pts = np.float32([
        [0, 0],
        [target_w - 1, 0],
        [target_w - 1, target_h - 1],
        [0, target_h - 1]
    ])

    # --- compute perspective transform from resized -> destination quadrilateral ---
    M = cv2.getPerspectiveTransform(src_pts, dst_pts.astype(np.float32))

    # --- warp the resized processed image back onto full image canvas ---
    h_full, w_full = orig_img.shape[:2]
    restored = cv2.warpPerspective(resized, M, (w_full, h_full))

    # --- create mask of the warped area (single channel) ---
    mask_src = np.ones((target_h, target_w), dtype=np.uint8) * 255
    mask_warped = cv2.warpPerspective(mask_src, M, (w_full, h_full))
    mask_warped = (mask_warped > 0).astype(np.uint8) * 255  # binary mask

    # --- blend: remove original region and add restored ---
    inv_mask = cv2.bitwise_not(mask_warped)
    inv_mask_3c = cv2.merge([inv_mask, inv_mask, inv_mask])

    cleaned_original = cv2.bitwise_and(orig_img, inv_mask_3c)
    final = cv2.add(cleaned_original, restored)

    return final



def resize_image(img, size="Letter", dpi=300):
    """
    Resize image to A4 or Letter paper size.
    Image is stretched to fill the page (aspect ratio may change).

    Parameters:
        img   : numpy ndarray (OpenCV image)
        size  : "Letter" or "A4"
        dpi   : 150, 200, or 300

    Returns:
        resized image (numpy ndarray)
    """

    size = size.lower()

    paper_sizes = {
        "letter": {
            150: (1275, 1650),
            200: (1700, 2200),
            300: (2550, 3300)
        },
        "a4": {
            150: (1240, 1754),
            200: (1654, 2339),
            300: (2480, 3508)
        }
    }

    if size not in paper_sizes:
        raise ValueError("size must be 'Letter' or 'A4'")

    if dpi not in paper_sizes[size]:
        raise ValueError("dpi must be 150, 200, or 300")

    target_size = paper_sizes[size][dpi]

    return cv2.resize(img, target_size, interpolation=cv2.INTER_CUBIC)


import shutil
from pathlib import Path
from fastapi.staticfiles import StaticFiles
from fastapi import FastAPI

def create_fresh_folder(
    app: FastAPI,
    folder_path: Path,
    mount_path: str,
    mount_name: str
):
    if folder_path.exists():
        shutil.rmtree(folder_path)

    folder_path.mkdir(parents=True, exist_ok=True)

    app.mount(
        mount_path,
        StaticFiles(directory=folder_path),
        name=mount_name
    )

    print(f"Prepared folder: {folder_path} (mounted at {mount_path})")






import cv2
import numpy as np
import os

def estimate_scanned_page_orientation(image_path):
    """
    Detect rectangles in the page and estimate orientation using soft criteria.
    Returns rotation in degrees: 0, 90, 180, 270
    """
    img = cv2.imread(image_path)
    if img is None:
        print("Failed to load image:", image_path)
        return None

    height, width = img.shape[:2]
    gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

    # Threshold to binary (invert so objects become white)
    _, thresh = cv2.threshold(gray, 150, 255, cv2.THRESH_BINARY_INV)
    contours, _ = cv2.findContours(thresh, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

    candidate_rects = []

    # --- rectangle detection ---
    for cnt in contours:
        x, y, w_rect, h_rect = cv2.boundingRect(cnt)
        area = cv2.contourArea(cnt)
        aspect_ratio = w_rect / float(h_rect)

        # Ignore border-touching or nearly-full-image rectangles
        if x <= 5 or y <= 5 or x + w_rect >= width - 5 or y + h_rect >= height - 5:
            continue
        if w_rect >= 0.95 * width and h_rect >= 0.95 * height:
            continue

        # Keep rectangles with reasonable aspect ratio
        if 0.7 < aspect_ratio < 5.0:
            candidate_rects.append({
                "center": (x + w_rect//2, y + h_rect//2),
                "area": area
            })

    if not candidate_rects:
        print("No candidate rectangles found!")
        return None

    # Sort candidates by area descending
    candidate_rects.sort(key=lambda r: r["area"], reverse=True)

    # Filter top rectangles where smallest >= 0.25 * largest, up to 3
    top_rects = []
    for rect in candidate_rects:
        if len(top_rects) < 3:
            if not top_rects or rect["area"] >= 0.25 * top_rects[0]["area"]:
                top_rects.append(rect)

    if len(top_rects) < 1:
        print("Not enough rectangles for orientation inference.")
        return None

    # --- orientation inference ---
    rotations = [0, 90, 180, 270]
    confidences = []

    for rot in rotations:
        score = 0.0

        for r in top_rects:
            cx, cy = r["center"]

            # Rotate coordinates virtually
            if rot == 0:
                rx, ry = cx, cy
            elif rot == 90:
                rx, ry = height - cy, cx
            elif rot == 180:
                rx, ry = width - cx, height - cy
            elif rot == 270:
                rx, ry = cy, width - cx

            # Soft criterion: closer to top -> higher score
            score += 1 - (ry / height)
        
        # Soft criterion: relative horizontal order (largest rectangle to right)
        if len(top_rects) > 1:
            # Compare smallest and largest
            small = min(top_rects, key=lambda r: r['area'])
            large = max(top_rects, key=lambda r: r['area'])
            scx, scy = small['center']
            lcx, lcy = large['center']
            # Rotate points
            if rot == 0:
                rsx, rsY = scx, scy
                rlx, rly = lcx, lcy
            elif rot == 90:
                rsx, rsY = height - scy, scx
                rlx, rly = height - lcy, lcx
            elif rot == 180:
                rsx, rsY = width - scx, height - scy
                rlx, rly = width - lcx, height - lcy
            elif rot == 270:
                rsx, rsY = scy, width - scx
                rlx, rly = lcy, width - lcx

            score += max(0, (rlx - rsx) / width)

        confidences.append(score)

    best_rot = rotations[np.argmax(confidences)]
    return best_rot


def estimate_scanned_page_orientation_long_omr(image_or_path):
    """
    Accepts either:
      - image path (str / Path)
      - OpenCV image (numpy ndarray)

    Returns:
      0 if largest rectangle is at bottom
      180 otherwise
    """

    # --- Load image if a path is provided ---
    if isinstance(image_or_path, (str, os.PathLike)):
        img = cv2.imread(str(image_or_path))
        if img is None:
            print("Failed to load image:", image_or_path)
            return None

    # --- Image already provided ---
    elif isinstance(image_or_path, np.ndarray):
        img = image_or_path.copy()

    else:
        raise TypeError(
            "image_or_path must be a file path or a numpy ndarray (OpenCV image)"
        )

    height, width = img.shape[:2]
    gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

    # Threshold to binary (invert so objects become white)
    _, thresh = cv2.threshold(gray, 150, 255, cv2.THRESH_BINARY_INV)
    contours, _ = cv2.findContours(
        thresh, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
    )

    candidate_rects = []

    for cnt in contours:
        x, y, w_rect, h_rect = cv2.boundingRect(cnt)
        area = cv2.contourArea(cnt)

        # Ignore border-touching or nearly-full-image rectangles
        if x <= 5 or y <= 5 or x + w_rect >= width - 5 or y + h_rect >= height - 5:
            continue
        if w_rect >= 0.95 * width and h_rect >= 0.95 * height:
            continue

        candidate_rects.append({
            "center": (x + w_rect // 2, y + h_rect // 2),
            "area": area
        })

    if not candidate_rects:
        print("No candidate rectangles found!")
        return None

    largest_rect = max(candidate_rects, key=lambda r: r["area"])
    _, cy = largest_rect["center"]

    return 0 if cy > height / 2 else 180



# -----------------------------
# Example usage
# -----------------------------
#img_path = '1.jpg'
#rotation = estimate_page_rotation(img_path)
#print(f"Estimated page rotation: {rotation} degrees")


from fastapi import Query,HTTPException
from fastapi.responses import JSONResponse
from pydantic import BaseModel
from typing import List, Optional
import os
import re
from pathlib import Path

#STATIC_DIR = "static"  # Make sure this matches the main.py path
documents_folder = Path.home() / "Documents"
question_bank_folder = documents_folder / "WebCam_OMR" / "Question_Bank"



async def where_to_copy_question_bank(base_folder: Path):
    """
    Returns the directory path where question bank files should be copied.
    Automatically creates the folder if it doesn't exist.
    
    Args:
        base_folder (Path): Base directory under which 'Question_Bank' folder will be created.
    
    Returns:
        dict: {
            "path": str(path_to_question_bank_folder),
            "example_file_name": example_file_name
        }
    """
    try:
        question_bank_folder = base_folder / "Question_Bank"

        # Ensure folder exists
        question_bank_folder.mkdir(parents=True, exist_ok=True)

        example_file_name = "CourseCode.tex  (e.g., EE3200.tex, EE1103.tex, CSE2229.tex etc.)"

        return {
            "path": str(question_bank_folder),
            "example_file_name": example_file_name
        }

    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Error creating folder: {e}")




# ------------------ Pydantic model ------------------
class AddQuestionRequest(BaseModel):
    department: str
    courseCode: str
    questionType: str
    questionStatement: str
    correctAns: str
    options: Optional[List[str]] = None
    mcqMark: Optional[float] = None


# ------------------ Load or create question bank ------------------
async def load_or_create_question_bank(base_folder: Path, courseCode: str = Query(...)):
    filename = f"{courseCode}.tex"
    question_bank_folder = base_folder / "Question_Bank"
    filepath = os.path.join(question_bank_folder, filename)

    file_created = False

    # If file does not exist, create it
    if not os.path.exists(filepath):
        file_created = True
        with open(filepath, "w", encoding="utf-8") as f:
            f.write(
                "%----------------------------------------------------\n"
                "%                  T/F questions\n"
                "%----------------------------------------------------\n\n"
                "%----------------------------------------------------\n"
                "%                  MCQ questions\n"
                "%----------------------------------------------------\n\n"
            )

    # Read the file content
    with open(filepath, "r", encoding="utf-8") as f:
        content = f.read()

    # Return JSON including content and optional file path if created
    response_data = {"texContent": content}
    if file_created:
        response_data["createdFilePath"] = filepath

    return JSONResponse(response_data)
    



# ------------------ Add question ------------------
async def add_question(base_folder: Path, request_add_q_data: AddQuestionRequest):
    data = request_add_q_data
    question_bank_folder = base_folder / "Question_Bank"

    filename = f"{data.courseCode}.tex"
    filepath = os.path.join(question_bank_folder, filename)

    if not os.path.exists(filepath):
        return JSONResponse(status_code=404, content={"error": "File not found"})

    # Read file lines
    with open(filepath, "r", encoding="utf-8") as f:
        lines = f.readlines()

    # Prepare new question text
    if data.questionType.lower() == "tf":
        print("tf question type detected...")
        answer = data.correctAns
#        answer = data.options[0] if data.options else "T"
        mark = 1
        new_question_lines = [
            f'\\question[type=tf, answer={answer}, mark={mark}]\n',
            f'{data.questionStatement}\n',
            "\n"
        ]
        # Find first T/F question
        insert_idx = next((i for i, line in enumerate(lines) if line.startswith(r"\question[type=tf")), len(lines))
    
    elif data.questionType.lower() == "mcq":
        print("MCQ type detected...")
        if not data.options or len(data.options) < 2:
            return JSONResponse(status_code=400, content={"error": "MCQ must have at least 2 options"})
        
        print("JSON response contains >=2 options")
        
        correct_answer = data.correctAns
#        correct_answer = data.options[0]
        choices_text = "\n".join([f"\\choice {opt}" for opt in data.options])
        mark = data.mcqMark or 1
        print("Data values:", correct_answer,"\n",choices_text,"\n",mark )

        new_question_lines = [
            f'\\question[type=mcq, answer={correct_answer}, mark={mark}]\n',
            f'{data.questionStatement}\n',
            '\\begin{choices}\n',
            f'{choices_text}\n',
            '\\end{choices}\n',
            "\n"
        ]
        
        print("new_question_lines are created....\n\n",new_question_lines)
        # Find first MCQ question
        insert_idx = next((i for i, line in enumerate(lines) if line.startswith(r"\question[type=mcq")), len(lines))
    
    else:
        return JSONResponse(status_code=400, content={"error": "Unknown question type"})

    # Insert the new question
    lines[insert_idx:insert_idx] = new_question_lines

    # Write updated content back
    with open(filepath, "w", encoding="utf-8") as f:
        f.writelines(lines)

    return {"status": "success", "message": f"{data.questionType} question added successfully."}






# ------------------ Delete questions ------------------
async def delete_questions(base_folder: Path, data: dict):
    department = data.get("department")
    course_code = data.get("courseCode")
    qtype_to_delete = data.get("type")
    ids_to_delete = set(map(int, data.get("questionIds", [])))


    question_bank_folder = base_folder / "Question_Bank"
    filename = f"{course_code}.tex"
    filepath = os.path.join(question_bank_folder, filename)
    

    if not os.path.exists(filepath):
        return JSONResponse({"error": "File not found"}, status_code=404)

    with open(filepath, "r", encoding="utf-8") as f:
        lines = f.readlines()

    new_lines = []
    tf_counter = 0
    mcq_counter = 0
    skip_block = False
    for line in lines:
        if line.startswith(r"\question"):
            type_match = re.search(r"type=(\w+)", line)
            qtype = type_match.group(1) if type_match else None

            if qtype == "tf":
                tf_counter += 1
                qid = tf_counter
            elif qtype == "mcq":
                mcq_counter += 1
                qid = mcq_counter
            else:
                qid = None

            skip_block = (qtype == qtype_to_delete and qid in ids_to_delete)

        if not skip_block:
            new_lines.append(line)

    with open(filepath, "w", encoding="utf-8") as f:
        f.writelines(new_lines)

    return JSONResponse({"message": f"Deleted {len(ids_to_delete)} question(s)"})



from setuptools import setup, find_packages

setup(
    name='jhOMR-webInterface',
    version='0.2.1',
    packages=find_packages(),  # Automatically find packages
    description='This is WebcamOMR package that enables creating three sets of questions by shuffling Set_1.docx. The shuffling will be done so that there will be minimum amount of overlap in the correct answers between any two sets. The package also enables checking exam scripts.',
    author='Md Nur Kutubul Alam',
    author_email='alamjhilam@gmail.com',
    classifiers=[
        'Programming Language :: Python :: 3',
         ],
    install_requires=[
        "fastapi",
        "uvicorn",                  # Used for running FastAPI applications
        "pandas",
        "numpy",
        "opencv-python",            # For cv2
        "python-multipart",         # FastAPI dependency for form/file uploads
        "python-docx",              # For docx
        "openpyxl",                 # For Excel
        "requests",                 # For HTTP requests
        "Jinja2",                   # For templates
        "starlette",                # FastAPI dependency, usually handled by FastAPI install
        "pydantic"                  # For BaseModel
    ]
)
